/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include <assert.h>
#include "lw.h"
using namespace lwlib;

namespace lwlib
{
	static bool Csg_NodeIsParent[Csg_NodeType::Last + 1] =
	{
		false, // LeafEmpty
		true,  // ParentUnion
		true,  // ParentIntersection
		true,  // ParentNegate
		false, // LeafHalfSpace
		false, // LeafSphere
	};

	static int Csg_NodeNumChildNodes[Csg_NodeType::Last + 1] =
	{
		0,  // LeafEmpty
		2,  // ParentUnion
		2,  // ParentIntersection
		1,  // ParentNegate
		0,  // LeafHalfSpace
		0,  // LeafSphere
	};

	lwlib_EnumStr
	(
		Csg_NodeType,
		"LeafEmpty",
		"ParentUnion",
		"ParentIntersection",
		"ParentNegate",
		"LeafHalfSpace",
		"LeafSphere",
	);

	static Csg_Node_t *Csg_NodeCopy(Csg_Node_t *node)
	{
		Csg_Node_t *newNode;

		newNode = NULL;

		if (node != NULL)
		{
			newNode = lwlib_CallocSingle(Csg_Node_t);
			*newNode = *node;

			if (Csg_NodeIsParent[node->type])
			{
				lwlib_RepeatArrayN(node->childNodes, 
					(size_t)Csg_NodeNumChildNodes[node->type], 
					i, childNode,
					newNode->childNodes[i] = Csg_NodeCopy(childNode)
					);
			}
		}

		return newNode;
	}

	static void Csg_NodeFree(Csg_Node_t *node)
	{
		if (node != NULL)
		{
			if (Csg_NodeIsParent[node->type])
			{
				lwlib_RepeatArrayNRev(node->childNodes, Csg_NodeNumChildNodes[node->type], 
					i, childNode,
					Csg_NodeFree(childNode)
					);
			}

			free(node);
		}
	}

	static bool Csg_NodeTestPoint(Csg_Node_t *node, Point3f point)
	{
		if (node != NULL)
		{
			if (node->type == Csg_NodeType::ParentIntersection)
			{
				return Csg_NodeTestPoint(node->childNodes[0], point) &&
					Csg_NodeTestPoint(node->childNodes[1], point);
			}
			if (node->type == Csg_NodeType::ParentUnion)
			{
				return Csg_NodeTestPoint(node->childNodes[0], point) ||
					Csg_NodeTestPoint(node->childNodes[1], point);
			}
			if (node->type == Csg_NodeType::ParentNegate)
			{
				return !Csg_NodeTestPoint(node->childNodes[0], point);
			}
			if (node->type == Csg_NodeType::LeafHalfSpace)
			{
				return vec_side(node->plane, point, 0.0) < 0;
			}
			if (node->type == Csg_NodeType::LeafSphere)
			{
				return vec3f_sphere_side(node->sphere, point, 0.0) < 0;
			}
		}

		return false;
	}

	static void Csg_NodeAddToEdit(Csg_Node_t *node, Edit *edit)
	{
		int i;
		Csg_Node_t *childNode;
		lwlib_String_t tmpStr;

		if (node != NULL)
		{
			edit->addItem_enum("type", (int *)&node->editType, 
				Csg_NodeType_Str, NULL, lwlib_CountArray(Csg_NodeType_Str));

			if (node->editType != node->type)
			{
				if (Csg_NodeIsParent[node->type])
				{
					lwlib_RepeatArrayN(node->childNodes, 
						(size_t)Csg_NodeNumChildNodes[node->type], 
						i, childNode, Csg_NodeFree(childNode));
					lwlib_Zero(node->childNodes);
				}

				if (Csg_NodeIsParent[node->editType])
				{
					lwlib_RepeatArrayN(node->childNodes,
						(size_t)Csg_NodeNumChildNodes[node->editType], 
						i, childNode, childNode = lwlib_CallocSingle(Csg_Node_t));
				}
				node->type = node->editType;
			}

			if (node->type == Csg_NodeType::LeafHalfSpace)
			{
				edit->addItem_TPlane3f("plane", &node->plane);
			}
			else if (node->type == Csg_NodeType::LeafSphere)
			{
				edit->addItem_TSphere3f("sphere", &node->sphere);
			}

			if (Csg_NodeIsParent[node->type])
			{
				for (i = 0; i < Csg_NodeNumChildNodes[node->type]; i++)
				{
					childNode = node->childNodes[i];

					lwlib_VaString(tmpStr, "childNode%d", i);
					lwlib_Edit_AddGroup(edit, tmpStr, 
						Csg_NodeAddToEdit(childNode, edit));
				}
			}
		}
	}

	static void Csg_NodeRotate(Csg_Node_t *node, Mat3f rotMat)
	{
		int i;
		Csg_Node_t *childNode;

		if (Csg_NodeIsParent[node->type])
		{
			for (i = 0; i < Csg_NodeNumChildNodes[node->type]; i++)
			{
				childNode = node->childNodes[i];
				Csg_NodeRotate(childNode, rotMat);
			}
		}
		else if (node->type == Csg_NodeType::LeafHalfSpace)
		{
			node->plane = mat3f_rotate_plane(rotMat, node->plane);
		}
		else if (node->type == Csg_NodeType::LeafSphere)
		{
			node->sphere.origin = mat_xform(rotMat, node->sphere.origin);
		}
	}

	static void Csg_NodeTranslate(Csg_Node_t *node, Vector3f xlateVec)
	{
		int i;
		Csg_Node_t *childNode;

		if (Csg_NodeIsParent[node->type])
		{
			for (i = 0; i < Csg_NodeNumChildNodes[node->type]; i++)
			{
				childNode = node->childNodes[i];
				Csg_NodeTranslate(childNode, xlateVec);
			}
		}
		else if (node->type == Csg_NodeType::LeafHalfSpace)
		{
			lwlib_plane_origin(node->plane) = vec_add(lwlib_plane_origin(node->plane), xlateVec);
		}
		else if (node->type == Csg_NodeType::LeafSphere)
		{
			node->sphere.origin = vec_add(node->sphere.origin, xlateVec);
		}
	}

	bool Csg_NodeTestBox(Csg_Node_t *node, Box3f box, OverlapTest::e test)
	{
		int i;
		bool result;
		bool childResult[2];
		Csg_Node_t *childNode;

		result = false;

		if (Csg_NodeIsParent[node->type])
		{
			if (node->type == Csg_NodeType::ParentNegate)
			{
				test = (OverlapTest::e)!test;
			}
			for (i = 0; i < Csg_NodeNumChildNodes[node->type]; i++)
			{
				childNode = node->childNodes[i];
				childResult[i] = Csg_NodeTestBox(childNode, box, test);
			}
			if (node->type == Csg_NodeType::ParentIntersection)
			{
				result = childResult[0] && childResult[1];
			}
			else if (node->type == Csg_NodeType::ParentUnion)
			{
				result = childResult[0] || childResult[1];
			}
			else if (node->type == Csg_NodeType::ParentNegate)
			{
				result = !childResult[0];
			}
		}
		else if (node->type == Csg_NodeType::LeafHalfSpace)
		{
			result = vec3f_box_halfspace_overlap(box, node->plane, test);
		}
		else if (node->type == Csg_NodeType::LeafSphere)
		{
			result = vec3f_box_sphere_overlap(box, node->sphere, test);
		}

		return result;
	}
}

Csg::Csg(Csg_Node_t *topNode)
{
	lwlib_Zero(priv);
	priv.topNode = topNode;
}

Csg::Csg()
{
	lwlib_Zero(priv);
	priv.topNode = lwlib_CallocSingle(Csg_Node_t);
}

Csg::~Csg()
{
	Csg_NodeFree(priv.topNode);
}

Csg *Csg::copy()
{
	Csg *csgCopy;
	Csg_Node_t *topNode;
	
	topNode = Csg_NodeCopy(priv.topNode);
	csgCopy = new Csg(topNode);

	return csgCopy;
}

Csg *Csg::intersect(Csg *a, Csg *b)
{
	Csg_Node_t *node;

	node = lwlib_CallocSingle(Csg_Node_t);
	node->type = Csg_NodeType::ParentIntersection;

	node->childNodes[0] = Csg_NodeCopy(a->priv.topNode);
	node->childNodes[1] = Csg_NodeCopy(b->priv.topNode);

	return new Csg(node);
}

Csg *Csg::csgUnion(Csg *a, Csg *b)
{
	Csg_Node_t *node;

	node = lwlib_CallocSingle(Csg_Node_t);
	node->type = Csg_NodeType::ParentUnion;

	node->childNodes[0] = Csg_NodeCopy(a->priv.topNode);
	node->childNodes[1] = Csg_NodeCopy(b->priv.topNode);

	return new Csg(node);
}

Csg *Csg::negate()
{
	Csg_Node_t *node;

	node = lwlib_CallocSingle(Csg_Node_t);
	node->type = Csg_NodeType::ParentNegate;

	node->childNodes[0] = Csg_NodeCopy(priv.topNode);

	return new Csg(node);
}

Csg *Csg::halfSpace(Plane3f plane)
{
	Csg_Node_t *node;

	node = lwlib_CallocSingle(Csg_Node_t);
	node->type = Csg_NodeType::LeafHalfSpace;
	node->plane = plane;

	return new Csg(node);
}

Csg *Csg::sphere(Sphere3f sphere)
{
	Csg_Node_t *node;

	node = lwlib_CallocSingle(Csg_Node_t);
	node->type = Csg_NodeType::LeafSphere;
	node->sphere = sphere;

	return new Csg(node);
}

void Csg::rotate(Mat3f rotMat)
{
	Csg_NodeRotate(priv.topNode, rotMat);
}

void Csg::translate(Vector3f xlateVec)
{
	Csg_NodeTranslate(priv.topNode, xlateVec);
}

bool Csg::testPoint(Point3f point)
{
	return Csg_NodeTestPoint(priv.topNode, point);
}

void Csg::addToEdit(Edit *edit)
{
	lwlib_Edit_AddGroup(edit, "topNode", 
		Csg_NodeAddToEdit(priv.topNode, edit));

	lwlib_Edit_CheckPersist(priv.persist, addToEdit(fsEdit));
}

void Csg::updateFrame()
{
}

bool Csg::testBox(Box3f box, OverlapTest::e test)
{
	return Csg_NodeTestBox(priv.topNode, box, test);
}
