/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_EDIT_H
#define LWLIB_EDIT_H

#include "lw_vec.h"
#include "lw_misc.h"
#include "lw_fs.h"
#include "lw_font.h"
#include "lw_keys.h"
#include "lw_timer.h"
#include "lw_strmap.h"
#include "vec.h"

#define lwlib_Edit_CheckPersist(persist, fncall) \
do { \
	if ((persist).commit) \
	{ \
		lwlib::Edit *fsEdit; \
		(persist).commit = false; \
 \
		fsEdit = new lwlib::Edit(); \
		fsEdit->newFs((persist).fileName, (persist).fileMode); \
		fncall; \
		delete fsEdit; \
	} \
} while(0)

#define lwlib_Edit_AddGroup(edit, itemName, fncall) \
do { \
	edit->addItem_group(itemName); \
	fncall; \
	edit->navigateBack(); \
} while(0)

#define lwlib_Edit_AddVec(edit, itemName, vec, \
	elemInit, elemCleanup, elemAddToEdit) \
	edit->addItem_vec(itemName, \
		(lwlib_VoidVec_t *)&(vec), \
		sizeof(vec_at(vec, 0)), \
		elemInit, elemCleanup, elemAddToEdit \
		)

namespace lwlib
{
	class Edit;

	typedef void (*VecInit_t)(void *elem);
	typedef void (*VecCleanup_t)(void *elem);
	typedef void (*VecAddToEdit_t)(void *elem, Edit *edit);

	namespace Edit_ItemFlag
	{
		enum e
		{
			ReadOnly,
			GroupExpand,
			Disabled,
			VecChange,
		};
	}

	typedef struct Edit_Persist_s
	{
		lwlib_String_t fileName;
		Fs_FileMode::e fileMode;
		bool commit;
	} Edit_Persist_t;

	namespace Edit_EditMode
	{
		enum e
		{
			Normal,
			String,
		};
	}

	namespace Edit_ItemType
	{
		enum e
		{
			Int,
			Short,
			UChar,
			Double,
			String,
			Bool,
			Enum,
			Group,
		};
	}

	namespace Edit_ItemGroupType
	{
		enum e
		{
			Normal,
			TSphere3f,
			TPlane3f,
			TPoint3f,
			TPoint2f,
			TPoint2i,
			TBox3f,
			Vec,
			VecElem,
		};
	}

	typedef struct Edit_Vec_s
	{
		lwlib_VoidVec_t *core;
		int elemSize;
		VecInit_t elemInit;
		VecCleanup_t elemCleanup;
		VecAddToEdit_t elemAddToEdit;
		int selector;
		bool doInsert;
		bool doDelete;
		bool insertCopy;
	} Edit_Vec_t;

	struct Edit_Item_s;
	typedef struct Edit_Item_s *Edit_ItemHandle_t;
	typedef vec_t(Edit_ItemHandle_t) Edit_ItemHandles_t;

	typedef struct Edit_Item_s
	{
		char name[lwlib_StrMap_MaxString];
		Edit_ItemType::e type;
		struct Edit_Item_s *parentItem;
		int flags; 
		int itemGroupIndex;
		lwlib_Byte_t *valueData;
		int valueLength;
		// Edit_ItemType(Int)
		int *intValue;
		// Edit_ItemType(Short)
		short *shortValue;
		// Edit_ItemType(UChar)
		unsigned char *ucharValue;
		// Edit_ItemType(Double)
		double *doubleValue;
		// Edit_ItemType(String)
		char *stringValue;
		int stringSize;
		// Edit_ItemType(Bool)
		bool *boolValue;
		// Edit_ItemType(Enum)
		int *enumValue;
		const char **enumValueNames;
		const int *enumValues;
		int enumNumValues;
		// Edit_ItemType(Group)
		Edit_ItemGroupType::e groupType;
		int groupItemCount;
		Edit_ItemHandles_t groupItems;
		Edit_Vec_t *groupVec;
		// Edit_ItemGroupType(TSphere3f)
		Sphere3f *tSphere3fValue;
		// Edit_ItemGroupType(TPlane3f)
		Plane3f *tPlane3fValue;
		// Edit_ItemGroupType(TPoint3f)
		Point3f *tPoint3fValue;
		// Edit_ItemGroupType(TPoint2f)
		Point2f *tPoint2fValue;
		// Edit_ItemGroupType(TPoint2i)
		Point2i *tPoint2iValue;
		// Edit_ItemGroupType(TBox3f)
		Box3f *tBox3fValue;
	} Edit_Item_t;

	typedef struct Edit_CopyRegister_s
	{
		Edit_ItemType::e type;
		lwlib_Byte_t *valueData;
		int valueLength;
	} Edit_CopyRegister_t;

	class Edit
	{
		struct
		{
			Timer timeNow;
			Edit_Item_t *topItem;
			Edit_Item_t *nav;
			vec_t(lwlib_Str_t) groupItemPath;
			lwlib_String_t selectedItemName;
			bool active;
			Timer cursorDrawTimer;
			int cursorCol;
			Fs *fs;
			bool fsOwner;
			Edit_EditMode::e editMode;
			Font *font;
			lwlib_Keys_t keys;
			Edit_CopyRegister_t copyRegisters[10];
			int scroll;
			Point2i selectedRange;
			bool gotSelectedRange;
			Point2i titleRange;
		} priv;

		Edit_Item_t *findItem(const char *itemName);

		Edit_Item_t *addItem(const char *itemName, 
			Edit_ItemType::e itemType);

		Edit_Item_t *addItem_WithValue(const char *itemName, 
			Edit_ItemType::e itemType, void *valueData,
			int valueLength);

		void resetItemSelectors();

		bool navigateItemByPath(lwlib_StrSeq_t itemPath);

		void item_DisplayString(Edit_Item_t *item, 
			lwlib_String_t dispStr);

		bool item_Selected(Edit_Item_t *item);

		int item_CursorCol(Edit_Item_t *item);

		void drawCursor(Edit_Item_t *item, Point2i pos);

		int queryIncDecKeys();

		bool item_HandleCopyPasteKeys(Edit_Item_t *item);

		void item_HandleIntEditKeys(Edit_Item_t *item);

		void item_HandleShortEditKeys(Edit_Item_t *item);

		void item_HandleUCharEditKeys(Edit_Item_t *item);

		void item_HandleDoubleEditKeys(Edit_Item_t *item);

		void item_HandleStringEditKeys(Edit_Item_t *item);

		void item_HandleGroupEditKeys(Edit_Item_t *item);

		void vecChangeNotify();

		int navItemDepth();

		const char *navItemDepthSpacing(lwlib_String_t depthStr);

		Point2i drawGroupItems(Edit_Item_t *item, Point2i pos);

		void cycleSelectedItem(int cycle);

		void scrollToSelected();

	public:
		Edit();

		~Edit();

		void newFs(const char *fileName, Fs_FileMode::e fileMode);

		void navigateTop();

		bool navigateItem(const char *itemName);

		bool navigateBack();

		void addItem_int(const char *itemName, int *value);

		void addItem_short(const char *itemName, short *value);

		void addItem_uchar(const char *itemName, unsigned char *value);

		void addItem_double(const char *itemName, double *value);

		void addItem_string(const char *itemName, char *value, 
			int valueSize);

		void addItem_bool(const char *itemName, bool *value);

		void addItem_enum(const char *itemName, int *value, 
			const char *valueNames[], const int *values, int numValues);

		void addItem_group(const char *itemName);

		void addItem_vec(const char *itemName, lwlib_VoidVec_t *value, 
			int elemSize, VecInit_t elemInit, VecCleanup_t elemCleanup, 
			VecAddToEdit_t elemAddToEdit);

		void addItem_TPoint3f(const char *itemName, Point3f *value);

		void addItem_TPoint2f(const char *itemName, Point2f *value);

		void addItem_TPoint2i(const char *itemName, Point2i *value);

		void addItem_TPlane3f(const char *itemName, Plane3f *value);

		void addItem_TBox3f(const char *itemName, Box3f *value);

		void addItem_TSphere3f(const char *itemName, Sphere3f *value);

		void addItem_persist(Edit_Persist_t *persist);

		void clearItems();

		int getItemFlags(const char *itemName);

		void setItemFlags(const char *itemName, int itemFlags);

		void setItemFlag(const char *itemName, Edit_ItemFlag::e itemFlag,
			bool enabled);

		void draw();

		void updateFrame();

		void setActive(bool active);

		bool isActive();

		void setFs(Fs *fs);

		void setFont(Font *font);

		void setKeys(lwlib_Keys_t keys);

		void addPersistGroup(Edit_Persist_t *persist);

		void setVecInit(const char *itemName, VecInit_t elemInit);

		void setVecCleanup(const char *itemName, VecCleanup_t elemCleanup);

		void setVecAddToEdit(const char *itemName,
			VecAddToEdit_t elemAddToEdit);

		void setItemEnabled(const char *itemName, bool enabled);
	};
}

#endif
