/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include "lw_font.h"
#include "lw_img.h"
using namespace lwlib;

#define LWLIB_FONT_SW 16
#define LWLIB_FONT_SH 16

typedef lwlib_Byte_t byte;

void Font::printChar(int c, int cx, int cy, int sw, int sh, 
	Font_Color::e curColor)
{
	int iw, ih;
	Gl *gl;
	Point4b colorRgba;

	assert(c >= 0 && c < 256);
	gl = priv.gl;

	iw = priv.fontImg->getWidth() / 16;
	ih = priv.fontImg->getHeight() / 16;

	if (priv.charImg == NULL)
	{
		priv.charImg = new Img();
	}

	priv.fontImg->refreshTex(gl);
	priv.fontImg->subImage(priv.charImg, (c % 16) * iw, 
		(c / 16) * ih, iw, ih);
	
	const ColorRemap_Color::e remapCurColor = 
		(ColorRemap_Color::e)curColor;
	priv.colorRemap->setActiveColor(remapCurColor);

	colorRgba = ColorRemap_GetColorRgba(remapCurColor);
	gl->setCurColor(colorRgba);

	priv.charImg->drawScaled(gl, cx, cy, sw, sh, true);

	gl->resetCurColor();
}

Point2i Font::printPosFitNoNl(Point2i pos, const char *text)
{
	int x, y;
	int cols;
	int n, r;
	const char *p;
	char *tmpStr;
	lwlib_String_t tmpStrNoAlloc;
	bool alloc;

	assert(priv.width > 0 && text != NULL);

	cols = priv.gl->getWidth() / priv.width;
	assert(priv.indent >= 0 && priv.indent < cols);

	x = lwlib_X(pos);
	y = lwlib_Y(pos);
	n = strlen(text);
	p = text;
	alloc = false;

	tmpStr = tmpStrNoAlloc;
	if ((size_t)cols > sizeof(tmpStrNoAlloc) - 1)
	{
		tmpStr = lwlib_CallocMany(char, cols + 1);
		alloc = true;
	}

	while (1)
	{
		if (x == cols)
		{
			x = priv.indent;
			y++;
		}

		r = cols - x;
		if (n > r)
		{
			memcpy(tmpStr, p, r);
			tmpStr[r] = '\0';
			printPos(x, y, tmpStr);

			x += r;
			n -= r;
			p += r;
		}
		else // n <= r
		{
			printPos(x, y, p);
			x += n;
			break;
		}
	}

	if (alloc)
	{
		free(tmpStr);
	}

	return vec2i(x, y);
}

Font::Font(const char *fontFile, SDL_Color *palette)
{
	lwlib_Zero(priv);

	priv.width = LWLIB_FONT_SW;
	priv.height = LWLIB_FONT_SH;
	priv.color = Font_Color::Grey;

	priv.fontImg = Img::load(fontFile);
#ifndef SDL_TRUE_COLOR
	lwlib_Img_Palettize(priv.fontImg, palette, true, 0xff);
#endif

	assert(priv.fontImg != NULL && 
		priv.fontImg->getWidth() == 256 &&
		priv.fontImg->getHeight() == 256);
}

Font::~Font()
{
	delete priv.fontImg;
	delete priv.charImg;
}

void Font::setShadow(int shadow)
{
	priv.shadow = shadow;
}

void Font::setColor(Font_Color::e color)
{
	priv.color = color;
}

void Font::print(int x, int y, const char *text)
{
	int cx, cy, sw, sh, c;
	Font_Color::e curColor; // current color (since ^n can change it)

	if (priv.fontImg == NULL)
	{
		return;
	}

	cx = x;
	cy = y;
	sh = priv.height;	// scaled height

	curColor = priv.color;

	while ((c = *text++) != '\0')
	{
		c = (byte)c;

		if (c == '\n')
		{
			cx = x;
			cy += sh;
			continue;
		}
		if (c == '^' && *text >= '0' && *text <= '8')
		{
			if (*text == '8')
			{
				curColor = priv.color;
			}
			else
			{
				curColor = (Font_Color::e)(*text - '0');
			}

			text++;
			continue;
		}
		sw = priv.width;
		if (priv.shadow)
		{
		}
		printChar(c, cx, cy, sw, sh, curColor);
		cx += sw;
	}
}

void Font::printPos(int xl, int yl, const char *text)
{
	print(xl * priv.width, yl * priv.height, text);
}

void Font::printPosOff(int xl, int yl, int ox, int oy, const char *text)
{
	print(ox + xl * priv.width, oy + yl * priv.height, text);
}

Point2i Font::printPosFit(Point2i pos, const char *text)
{
	int x0;
	char *start, *p;
	lwlib_String_t tmpStr;

	assert(priv.width > 0 && text != NULL);

	lwlib_Snprintf(tmpStr, "%s", text);

	x0 = lwlib_X(pos);
	start = tmpStr;

	p = strchr(start, '\n');
	while (p != NULL)
	{
		*p = '\0';
		pos = printPosFitNoNl(pos, start);
		lwlib_X(pos) = x0;
		lwlib_Y(pos)++;

		start = p + 1;
		p = strchr(start, '\n');
	}

	return printPosFitNoNl(pos, start);
}

Point2i Font::advancePosFit(Point2i pos, int col)
{
	int x, y;
	int cols;
	int n, r;

	assert(priv.width > 0);

	cols = priv.gl->getWidth() / priv.width;
	assert(priv.indent >= 0 && priv.indent < cols);

	x = lwlib_X(pos);
	y = lwlib_Y(pos);
	n = col;

	if (col == -1)
	{
		x = cols;
		n = 0;
	}

	while (1)
	{
		if (x == cols)
		{
			x = priv.indent;
			y++;
		}

		r = cols - x;
		if (n > r)
		{
			x += r;
			n -= r;
		}
		else // n <= r
		{
			x += n;
			break;
		}
	}

	return vec2i(x, y);
}

void Font::setColorRemap(ColorRemap *colorRemap)
{
	priv.colorRemap = colorRemap;
}

void Font::setGl(Gl *gl)
{
	priv.gl = gl;
}

int Font::getIndent()
{
	return priv.indent;
}

void Font::setIndent(int indent)
{
	priv.indent = indent;
}

int Font::getWidth()
{
	return priv.width;
}

int Font::getHeight()
{
	return priv.height;
}

int Font::maxScreenPosY()
{
	return (priv.gl->getHeight() / LWLIB_FONT_SH) - 1;
}
