/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

//#define LWLIB_ENABLE_JPEG

#include "lw.h"
#include <errno.h>
#ifdef LWLIB_ENABLE_JPEG
#include <jpeglib.h>
#endif
using namespace lwlib;

#define byte lwlib_Byte_t
#define Img_MaxOsPath 128

typedef unsigned short word;

// ------------------------- * TGA (Truevision TARGA) * -------------------------
#pragma pack(push, 1)
typedef struct tga_s
{
	byte id_length, colormap_type, type;
	word colormap_index, colormap_length;
	byte colormap_size;
	word x_origin, y_origin, width, height;
	byte pixel_size, attributes;
} tga_t;
#pragma pack(pop)

bool Img::readTGA(char *name)
{
	byte *raw, *p, *dest;
	tga_t *header;
	byte pheader;
	bool flip;
	int plen, x, y0, y;
	Fs *fs;

	fs = new Fs(name, lwlib::Fs_FileMode::Read);
	if (!fs->loadEntireFile())
	{
		return false; // read error
	}

	raw = fs->getData();
	header = (tga_t *)raw;

	// check image type
	if ((header->type != 2 && header->type != 10) ||
		(header->pixel_size != 32 && header->pixel_size != 24)) 
	{
		delete fs;
		return false; // unsupported type
	}

	priv.w = header->width;
	priv.wshift = lwlib_Log2(priv.w);
	priv.h = header->height;
	priv.hshift = lwlib_Log2(priv.h);
	priv.bpp = (header->pixel_size) / 8;
	priv.pitch = priv.bpp * priv.w;
	priv.data = (byte *)malloc(priv.w * priv.h * priv.bpp);
	if (!priv.data)
	{
		delete fs;
		return false; // memory allocation error
	}
	priv.dataOwned = true;

	flip = !(header->attributes & 0x20); // top-bottom orientation
	p = raw + sizeof(tga_t) + header->id_length; // data block
	if (header->colormap_type)
	{
		p += (header->colormap_size / 8) * header->colormap_length; // skip colormap
	}

	switch (header->type) 
	{
	case 2: // Unmapped RGB images
		for (y0 = 0; y0 < priv.h; y0++)
		{
			y = flip ? (priv.h - (y0 + 1)) : y0;
			dest = priv.data + y * priv.w * priv.bpp;

			// copy scanline
			for (x = 0; x < priv.w; x++)
			{
				dest[2] = *p++;
				dest[1] = *p++;
				dest[0] = *p++;
				if (priv.bpp == 4)
					dest[3] = *p++;
				dest += priv.bpp;
			}
		}
		break;
	case 10: // Run Length Encoded, RGB images
		for (y0 = 0; y0 < priv.h; y0++)
		{
			y = flip ? (priv.h - (y0 + 1)) : y0;
			dest = priv.data + y * priv.w * priv.bpp;
			for (x = 0; x < priv.w;)
			{
				pheader = *p++;
				plen = (pheader & 127) + 1;

				if (pheader & 128) // RLE packet
				{
					while (plen--)
					{
						dest[2] = p[0];
						dest[1] = p[1];
						dest[0] = p[2];
						if (priv.bpp == 4)
							dest[3] = p[3];
						dest += priv.bpp;

						if (++x == priv.w) // run spans across rows
						{
							if (++y0 == priv.h)
								break; // last scanline decoded
							y = flip ? (priv.h - (y0 + 1)) : y0;
							dest = priv.data + y * priv.w * priv.bpp;
							x = 0;
						}
					}
					p += priv.bpp;
				}
				else
				{
					while (plen--)
					{
						dest[2] = *p++;
						dest[1] = *p++;
						dest[0] = *p++;
						if (priv.bpp == 4)
							dest[3] = *p++;
						dest += priv.bpp;

						if (++x == priv.w) // pixel packet run spans across rows
						{
							if (++y0 == priv.h)
								break; // last scanline decoded
							y = flip ? (priv.h - (y0 + 1)) : y0;
							dest = priv.data + y * priv.w * priv.bpp;
							x = 0;
						}
					}
				}
			}
		}
		break;
	}

	// TODO: flip as post-processing ?

	delete fs;

	initDefaultAttr();
	return true;
}

// ------------------------- * PCX (Zsoft Paintbrush) * -------------------------
#pragma pack(push, 1)
typedef struct pcx_s
{
	byte manufacturer;
	byte version;
	byte encoding;
	byte bits_per_pixel;
	word xmin, ymin, xmax, ymax; // w=xmax-xmin; h=ymax-ymin
	word hres, vres;
	byte palette[48];
	byte reserved;
	byte color_planes;
	word bytes_per_line;
	word palette_type;
	byte filler[58];
} pcx_t;
#pragma pack(pop)

bool Img::readPCX(char *name)
{
	byte *raw, *p, *dest, *palette;
	pcx_t *header;
	int data, len, x, y;
	Fs *fs;

	fs = new Fs(name, lwlib::Fs_FileMode::Read);
	if (!fs->loadEntireFile())
	{
		return false; // file read error
	}

	raw = fs->getData();
	header = (pcx_t *)raw;
	len = fs->getLength();

	// check image type
	if (header->manufacturer != 0x0A ||
		header->version != 5 ||
		header->encoding != 1 || header->bits_per_pixel != 8)
	{
		delete fs;
		return false; // unsupported type
	}

	priv.w = header->xmax - header->xmin + 1;
	priv.wshift = lwlib_Log2(priv.w);
	priv.h = header->ymax - header->ymin + 1;
	priv.hshift = lwlib_Log2(priv.h);
	priv.bpp = 3;
	priv.pitch = priv.bpp * priv.w;
	priv.data = (byte *)malloc(priv.w * priv.h * priv.bpp);
	if (!priv.data)
	{
		delete fs;
		return false; // memory allocation error
	}
	priv.dataOwned = true;

	palette = raw + len - 768;
	p = raw + sizeof(pcx_t);

	for (y = 0; y < priv.h; y++)
	{
		dest = priv.data + y * priv.w * priv.bpp;

		for (x = 0; x < priv.w;)
		{
			data = *p++;

			if ((data & 0xC0) == 0xC0)
			{
				len = data & 0x3F;
				data = *p++;
			} else
				len = 1;

			while (len--)
			{
				*dest++ = palette[data * 3 + 0];
				*dest++ = palette[data * 3 + 1];
				*dest++ = palette[data * 3 + 2];
				if (++x >= priv.w)
					break;
			}
		}
	}

	delete fs;

	initDefaultAttr();
	return true;
}

#ifdef LWLIB_ENABLE_JPEG
// ------------------------- * JPEG (Joint Photographic Experts Group) * -------------------------

static void jpg_noop(j_decompress_ptr cinfo)
{
}

static boolean jpg_fill_input_buffer(j_decompress_ptr cinfo)
{
	return true;
}

static void jpg_skip_input_data(j_decompress_ptr cinfo, long nupriv.bytes)
{
	cinfo->src->next_input_byte += (size_t)nupriv.bytes;
	cinfo->src->bytes_in_buffer -= (size_t)nupriv.bytes;
}

static void jpeg_mepriv.src(j_decompress_ptr cinfo, byte * mem, int len)
{
	cinfo->src = (struct jpeg_source_mgr *)
		(*cinfo->mem->alloc_small)(
			(j_common_ptr)cinfo, 
			JPOOL_PERMANENT, 
			sizeof(struct jpeg_source_mgr)
			);
	cinfo->src->init_source = jpg_noop;
	cinfo->src->fill_input_buffer = jpg_fill_input_buffer;
	cinfo->src->skip_input_data = jpg_skip_input_data;
	cinfo->src->resync_to_restart = jpeg_resync_to_restart;
	cinfo->src->terpriv.source = jpg_noop;
	cinfo->src->bytes_in_buffer = len;
	cinfo->src->next_input_byte = mem;
}

bool Img_ReadJPG(Img_t img, char *name)
{
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	JSAMPARRAY buffer;
	byte *raw, *dest;
	int len;
	lwlib_Fs_t fs;

	fs = lwlib_Fs_New(name, lwlib_Fs_FileMode(Read));
	if (!lwlib_Fs_LoadEntireFile(fs))
	{
		return false; // file read error
	}

	raw = (byte *)lwlib_Fs_GetData(fs);
	len = lwlib_Fs_GetLength(fs);

	// decompress
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);
	jpeg_mepriv.src(&cinfo, raw, len);
	jpeg_read_header(&cinfo, TRUE);
	jpeg_start_decompress(&cinfo);

	// check image type
	if (cinfo.output_components != 3)
	{
		jpeg_destroy_decompress(&cinfo);
		lwlib_Fs_Free(fs);
		return false; // unsupported type
	}

	priv.w = cinfo.output_width;
	priv.wshift = lwlib_Log2(priv.w);
	priv.h = cinfo.output_height;
	priv.hshift = lwlib_Log2(priv.h);
	priv.bpp = 3;
	priv.data = malloc(priv.w * priv.h * priv.bpp);
	if (!priv.data)
	{
		jpeg_destroy_decompress(&cinfo);
		lwlib_Fs_Free(fs);
		return false; // memory allocation error
	}
	priv.dataOwned = true;

	while (cinfo.output_scanline < cinfo.output_height)
	{
		dest = priv.data + cinfo.output_scanline * priv.w * priv.bpp;
		buffer = &dest;
		jpeg_read_scanlines(&cinfo, buffer, 1);
	}

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);
	lwlib_Fs_Free(fs);
	return true;
}
#endif

// ------------------------- * common * -------------------------

void Img::initDefaultAttr()
{
	Img_MipMapLevel_t *level;

	priv.mipmap.numLevels = 1;

	level = &priv.mipmap.levels[0];
	level->w = priv.w;
	level->h = priv.h;
	level->source = priv.data;
	level->wshift = lwlib_Log2(priv.w);
	level->hshift = lwlib_Log2(priv.h);
	level->pitch = priv.pitch;

	priv.sub.built = false;
	priv.sub.x = 0;
	priv.sub.y = 0;
	priv.sub.w = priv.w;
	priv.sub.h = priv.h;
	priv.sub.wshift = lwlib_Log2(priv.w);
	priv.sub.hshift = lwlib_Log2(priv.h);
	priv.sub.data = priv.data;
}

Img_t Img::load(const char *path)
{
	bool success = false;
	Img_t img;
	char name[Img_MaxOsPath], *src;

	lwlib_Zero(img);
	if (!path || !*path)
	{
		return img;
	}

	img = new Img();

	strcpy(name, path);

	// find extension
	src = name + strlen(name) - 1;
	while (*src != '/' && src != name)
	{
		if (*src == '.')
			break;
		src--;
	}

	if (*src == '.')
	{
		if (!lwlib_strcasecmp(src, ".tga"))
			success = img->readTGA(name);
		#ifdef LWLIB_ENABLE_JPEG
		else if (!lwlib_strcasecmp(src, ".jpg"))
			success = img->readJPG(name);
		#endif
		else if (!lwlib_strcasecmp(src, ".pcx"))
			success = img->readPCX(name);
	}
	else
	{
		src = name + strlen(name);
	
		strcpy(src, ".tga");
		if (!success && img->readTGA(name))
			success = true;
		#ifdef LWLIB_ENABLE_JPEG
		strcpy(src, ".jpg");
		if (!success && img->readJPG(name))
			success = true;
		#endif
		strcpy(src, ".pcx");
		if (!success && img->readPCX(name))
			success = true;
	}

	if (!success)
	{
		delete img;
	}

	return img;
}

void Img::fromSource(lwlib_Byte_t *source, int w, int h, 
	int bpp, int pitch, bool owned, SDL_Color *palette)
{
	priv.data = source;
	priv.w = w;
	priv.h = h;
	priv.wshift = lwlib_Log2(priv.w);
	priv.hshift = lwlib_Log2(priv.h);
	priv.bpp = bpp;
	priv.pitch = pitch;
	priv.dataOwned = owned;
	priv.mipmap.built = false;
	priv.palette = palette;

	initDefaultAttr();
}

lwlib_Byte_t Img_FindClosestColor(SDL_Color *palette, int r, 
	int g, int b)
{
	int i;
	byte col = 0;
	int dist, mindist;

	mindist = 256 * 256 * 3 + 1;
	for (i = 0; i < 256; i++)
	{
		dist =
			lwlib_Sq(palette->r - r) +
			lwlib_Sq(palette->g - g) +
			lwlib_Sq(palette->b - b);
		if (dist < mindist)
		{
			mindist = dist;
			col = (byte)i;
		}
		palette++;
	}

	return col;
}


Img::Img()
{
	lwlib_Zero(priv.w);
	lwlib_Zero(priv.h);
	lwlib_Zero(priv.wshift);
	lwlib_Zero(priv.hshift);
	lwlib_Zero(priv.bpp);
	lwlib_Zero(priv.pitch);
	lwlib_Zero(priv.data);
	lwlib_Zero(priv.mipmap);
	lwlib_Zero(priv.sub);
	lwlib_Zero(priv.dataOwned);
	lwlib_Zero(priv.texId);
	lwlib_Zero(priv.palette);
}

void Img::palettize(SDL_Color *palette, bool transparent, 
	lwlib_Byte_t transkey)
{
	int i;
	int newLen;
	byte *p, *pNew;
	byte *newData;
	byte col;
	Img_MipMapLevel_t *level;

	assert(priv.data != NULL && (priv.bpp == 1 || priv.bpp == 3 || 
		priv.bpp == 4) && !priv.mipmap.built && !priv.sub.built);

	if (priv.bpp == 1)
	{
		return;
	}

	newLen = priv.w * priv.h;
	newData = lwlib_CallocMany(byte, newLen);

	p = priv.data;
	pNew = newData;

	if (priv.bpp == 3 || (priv.bpp == 4 && !transparent))
	{
		for (i = 0; i < newLen; i++)
		{
			*pNew = Img_FindClosestColor(palette, (int)p[0], 
				(int)p[1], (int)p[2]);
			p += priv.bpp;
			pNew++;
		}
	}
	else
	{
		assert(priv.bpp == 4 && transparent);

		for (i = 0; i < newLen; i++)
		{
			if (p[3] == 0)
			{
				col = transkey;
			}
			else
			{
				const int r = (int)p[0] * (int)p[3] / 255;
				const int g = (int)p[1] * (int)p[3] / 255;
				const int b = (int)p[2] * (int)p[3] / 255;
				col = Img_FindClosestColor(palette, r, g, b);
			}
			*pNew = col;

			p += priv.bpp;
			pNew++;
		}
	}

	if (priv.dataOwned)
	{
		free(priv.data);
	}
	priv.data = newData;
	priv.dataOwned = true;
	priv.bpp = 1;
	priv.pitch = priv.w;

	level = &priv.mipmap.levels[0];
	level->pitch = priv.pitch;
	level->source = priv.data;

	priv.sub.data = priv.data;
}

Img::~Img()
{
	if (priv.dataOwned)
	{
		free(priv.data);
	}
}

int Img::getWidth()
{
	Img_MipMapLevel_t *level;

	if (priv.sub.built)
	{
		return priv.sub.w;
	}

	level = &priv.mipmap.levels[priv.mipmap.curLevel];
	return level->w;
}

int Img::getHeight()
{
	Img_MipMapLevel_t *level;

	if (priv.sub.built)
	{
		return priv.sub.h;
	}

	level = &priv.mipmap.levels[priv.mipmap.curLevel];
	return level->h;
}

int Img::getWidthShift()
{
	Img_MipMapLevel_t *level;

	if (priv.sub.built)
	{
		return priv.sub.wshift;
	}

	level = &priv.mipmap.levels[priv.mipmap.curLevel];
	return level->wshift;
}

int Img::getHeightShift()
{
	Img_MipMapLevel_t *level;

	if (priv.sub.built)
	{
		return priv.sub.hshift;
	}

	level = &priv.mipmap.levels[priv.mipmap.curLevel];
	return level->hshift;
}

int Img::getBpp()
{
	return priv.bpp;
}

int Img::getPitch()
{
	Img_MipMapLevel_t *level;

	level = &priv.mipmap.levels[priv.mipmap.curLevel];
	return level->pitch;
}

lwlib_Byte_t *Img::getData()
{
	Img_MipMapLevel_t *level;

	if (priv.sub.built)
	{
		return priv.sub.data;
	}

	level = &priv.mipmap.levels[priv.mipmap.curLevel];
	return level->source;
}

void Img::applyMonotone(lwlib_Byte_t monosrckey, lwlib_Byte_t monodestkey)
{
	int i;
	byte *p;

	assert(!priv.sub.built);

	if (priv.bpp != 1 || priv.data == NULL)
	{
		return;
	}

	p = priv.data;
	for (i = 0; i < priv.w * priv.h; i++)
	{
		if (*p != monosrckey)
		{
			*p = monodestkey;
		}
		p++;
	}
}

void Img::buildMipMaps(SDL_Color *palette)
{
	int i, j, x, y;
	int pixoff;
	int size, texsize, texshift;
	byte col;
	int r, g, b, a;
	byte *source, *newdata, *dest, *pix;
	Img_MipMapLevel_t *level;

	assert(priv.data != NULL && priv.w > 0 && priv.h > 0 && 
		!priv.mipmap.built && priv.w == priv.h && 
		lwlib_IntegerIsPow2(priv.w) && !priv.sub.built);
	
	texsize = priv.w;
	texshift = lwlib_Log2(texsize);

	size = 0;
	for (i = 0; i <= texshift; i++)
	{
		size += lwlib_Sq(1 << i);
	}
	newdata = lwlib_CallocMany(byte, size * priv.bpp);

	dest = newdata;
	source = priv.data;

	size = texsize * texsize * priv.bpp;
	memcpy(dest, source, size);
	dest += size;

	for (i = 0; i < texshift; i++)
	{
		for (x = 0; x < texsize; x += 2)
		{
			for (y = 0; y < texsize; y += 2)
			{
				if (priv.bpp == 1)
				{
					r = g = b = 0;
					for (j = 0; j < 4; j++)
					{
						col = source[(y + (j & 1)) + 
							((x + (j / 2)) << (texshift - i))];
						r += palette[col].r;
						g += palette[col].g;
						b += palette[col].b;
					}
					r >>= 2;
					g >>= 2;
					b >>= 2;

					col = Img_FindClosestColor(palette, r, g, b);
					dest[(y >> 1) + ((x >> 1) << (texshift - i - 1))] = col;
				}
				else
				{
					r = g = b = a = 0;
					for (j = 0; j < 4; j++)
					{
						pixoff = (y + (j & 1)) + 
							((x + (j / 2)) << (texshift - i));
						pix = &source[pixoff * priv.bpp];
						r += pix[0];
						g += pix[1];
						b += pix[2];
						if (priv.bpp == 4)
						{
							a += pix[3];
						}
					}

					pixoff = (y >> 1) + ((x >> 1) << (texshift - i - 1)); 
					pix = &dest[pixoff * priv.bpp];
					pix[0] = (byte)(r >> 2);
					pix[1] = (byte)(g >> 2);
					pix[2] = (byte)(b >> 2);
					if (priv.bpp == 4)
					{
						pix[3] = (byte)(a >> 2);
					}
				}
			}
		}

		source = dest;
		size >>= 2;
		texsize >>= 1;
		dest += size;
	}

	if (priv.dataOwned)
	{
		free(priv.data);
	}
	priv.data = newdata;
	priv.dataOwned = true;
	priv.mipmap.built = true;
	priv.mipmap.texsize = priv.w;
	priv.mipmap.texshift = texshift;
	priv.mipmap.curLevel = 0;
	priv.mipmap.numLevels = 0;

	source = newdata;
	for (i = 0; i < LWLIB_IMG_MIPMAPLEVEL_MAX; i++)
	{
		if (i > priv.mipmap.texshift)
		{
			break;
		}

		level = &priv.mipmap.levels[i];
		level->source = source;
		level->wshift = level->hshift = priv.mipmap.texshift - i;
		level->w = level->h = (1 << level->wshift);
		level->pitch = level->w * priv.bpp;

		size = lwlib_Sq(priv.mipmap.texsize) * priv.bpp;
		source += size;

		priv.mipmap.numLevels++;
	}

	priv.sub.data = newdata;
}

lwlib_Byte_t Img::getPix(fixedpt x, fixedpt y)
{
	int s, t;
	Img_MipMapLevel_t *level;

	assert(priv.mipmap.curLevel >= 0 && 
		priv.mipmap.curLevel < priv.mipmap.numLevels &&
		!priv.sub.built);

	level = &priv.mipmap.levels[priv.mipmap.curLevel];
	s = fixedpt_toint(fixedpt_fracpart(x) << level->wshift);
	t = fixedpt_toint(fixedpt_fracpart(y) << level->hshift);
	return level->source[(s << level->wshift) + t];
}

void Img::setMipLevel(int curLevel)
{
	priv.mipmap.curLevel = curLevel;
}

void Img::transpose()
{
	int x, y;
	byte *newData;
	int newW, newH;

	assert(priv.data != NULL && priv.bpp == 1 && 
		priv.w > 0 && priv.h > 0 && !priv.mipmap.built &&
		!priv.sub.built);

	newData = lwlib_CallocMany(byte, priv.w * priv.h);
	newW = priv.h;
	newH = priv.w;

	for (x = 0; x < priv.w; x++)
	{
		for (y = 0; y < priv.h; y++)
		{
			newData[x * newW + y] = priv.data[y * priv.w + x];
		}
	}

	if (priv.dataOwned)
	{
		free(priv.data);
	}
	priv.data = newData;
	priv.dataOwned = true;
	priv.w = newW;
	priv.h = newH;
	priv.wshift = lwlib_Log2(priv.w);
	priv.hshift = lwlib_Log2(priv.h);
	priv.pitch = newW;

	initDefaultAttr();
}

void Img::drawScaled(Gl *gl, int cx, int cy, 
	int sw, int sh, bool transparent)
{
	refreshTex(gl);
	gl->drawScaledImage(priv.texId, 
		priv.sub.x, priv.sub.y, priv.sub.w, priv.sub.h, 
		priv.w, priv.h, cx, cy, sw, sh, transparent);
}

void Img::subImage(Img_t subImg, int ix, int iy, int iw, int ih)
{
	assert(!priv.mipmap.built && ix >= 0 && iy >= 0 && iw >= 0 && ih >= 0 &&
		!priv.sub.built);

	if (ix + iw <= 0 || ix >= priv.w)
	{
		ix = 0;
		iw = priv.w;
	}
	if (iy + ih <= 0 || iy >= priv.h)
	{
		iy = 0;
		ih = priv.h;
	}

	if (ix < 0)
	{
		iw += ix;
		ix = 0;
	}
	if (ix + iw > priv.w)
	{
		iw = priv.w - ix;
	}

	if (iy < 0)
	{
		ih += iy;
		iy = 0;
	}
	if (iy + ih > priv.h)
	{
		ih = priv.h - iy;
	}

	subImg->priv = priv;
	subImg->priv.dataOwned = false;

	subImg->priv.sub.built = true;
	subImg->priv.sub.x = ix;
	subImg->priv.sub.y = iy;
	subImg->priv.sub.w = iw;
	subImg->priv.sub.h = ih;
	subImg->priv.sub.wshift = lwlib_Log2(iw);
	subImg->priv.sub.hshift = lwlib_Log2(ih);
	subImg->priv.sub.data = priv.data + (((iy * priv.pitch) + ix) * priv.bpp);
}

void Img::expandPalette(bool transparent)
{
	int x, y, w, h, srcp, size;
	byte *ssrc, *src, *dst, *data;
	int dstbpp;

	assert(priv.data != NULL && priv.bpp == 1 && 
		!priv.mipmap.built && !priv.sub.built &&
		priv.texId == 0);

	src = priv.data;
	w = priv.w;
	h = priv.h;

	ssrc = src;
	size = w * h;

	dstbpp = transparent ? 4 : 3;
	data = lwlib_CallocMany(byte, size * dstbpp);
	dst = data;

	for (y = 0; y < h; y++)
	{
		for (x = 0; x < w; x++)
		{
			srcp = *src++;
			if (transparent && srcp == 255)
			{
				int r, g, b, i, c;
				int u, v;

				r = g = b = c = 0;
				for (v = y - 1; v <= y + 1; v++)
				{
					for (u = x - 1; u <= x + 1; u++)
					{
						if (u > 0 && u < w && v > 0 && v < h
							&& (i = ssrc[v * w + u]) != 255)
						{
							r += priv.palette[i].r;
							g += priv.palette[i].g;
							b += priv.palette[i].b;
							c++;
						}
					}
				}
				if (!c)
					c = 1;
				*dst++ = (byte) (r / c);
				*dst++ = (byte) (g / c);
				*dst++ = (byte) (b / c);
				*dst++ = 0;
				continue;
			}
			*dst++ = priv.palette[srcp].r;
			*dst++ = priv.palette[srcp].g;
			*dst++ = priv.palette[srcp].b;
			if (transparent)
				*dst++ = 255;
		}
	}
	
	priv.bpp = dstbpp;
	priv.pitch = dstbpp * w;
	if (priv.dataOwned)
	{
		free(priv.data);
	}
	priv.data = data;
	priv.mipmap.levels[0].pitch = priv.pitch;
	priv.mipmap.levels[0].source = data;
	priv.dataOwned = true;
	priv.sub.data = data;
}

void Img::refreshTex(Gl *gl)
{

	if (priv.texId == 0)
	{
		priv.texId = gl->uploadTex(priv.data, priv.w, priv.h, 
			priv.bpp, priv.palette);
	}
}

int Img::texId()
{
	return priv.texId;
}

void Img::scaleIntensity(lwlib_Byte_t blend)
{
	int i, o, c;
	int x, y;
	byte *data;

	assert((priv.bpp == 1 || priv.bpp == 3 || priv.bpp == 4) && 
		priv.data != NULL && priv.w != 0 &&
		priv.h != 0 && priv.pitch != 0);
	
	if (priv.bpp == 1)
	{
		data = priv.data;
		for (y = 0; y < priv.h; y++)
		{
			for (x = 0; x < priv.w; x++)
			{
				if (*data == 0xFF)
				{
					// don't touch transparent pixels
				}
				else
				{
					*data = Img_FindClosestColor(priv.palette,
						priv.palette[*data].r * (255 - blend) / 255, 
						priv.palette[*data].g * (255 - blend) / 255, 
						priv.palette[*data].b * (255 - blend) / 255);
				}
				data++;
			}
		}
	}
	else
	{
		for (y = 0; y < priv.h; y++)
		{
			for (x = 0; x < priv.w; x++)
			{
				o = (y * priv.pitch) + (x * priv.bpp);
				for (i = 0; i < 3; i++)
				{
					c = (int)priv.data[o + i];
					c = (c * (int)(255 - blend)) / 255;
					priv.data[o + i] = (byte)c;
				}
			}
		}
	}
}

void Img::copyData()
{
	byte *data;

	assert(priv.data != NULL);
	assert(priv.bpp == 1 || priv.bpp == 3 || priv.bpp == 4);
	assert(priv.w > 0 && priv.h > 0);
	assert(!priv.mipmap.built);
	assert(!priv.sub.built);

	data = (byte *)lwlib_CallocCopy(priv.data, priv.h * priv.pitch);

	if (priv.dataOwned)
	{
		free(priv.data);
	}
	priv.data = data;
	priv.dataOwned = true;

	initDefaultAttr();
}

void Img::expandDimsPow2()
{
	int sw, sh;
	byte *data;

	assert(priv.data != NULL);
	assert(priv.bpp == 1 || priv.bpp == 3 || priv.bpp == 4);
	assert(priv.w > 0 && priv.h > 0);
	assert(!priv.mipmap.built);
	assert(!priv.sub.built);

	for (sw = 1; sw < priv.w; sw <<= 1);
	for (sh = 1; sh < priv.h; sh <<= 1);

	if (sw == priv.w && sh == priv.h)
	{
		return;
	}

	data = lwlib_CallocMany(byte, sw * sh * priv.bpp);
	if (priv.bpp == 1)
	{
		resamplePaletted(data, sw, sh, priv.data, priv.w, 
			priv.h);
	}
	else
	{
		assert(priv.bpp == 3 || priv.bpp == 4);
		resample(data, sw, sh, priv.data, priv.w, priv.h,
			priv.bpp);
	}

	if (priv.dataOwned)
	{
		free(priv.data);
	}
	priv.data = data;
	priv.dataOwned = true;
	priv.w = sw;
	priv.h = sh;
	priv.wshift = lwlib_Log2(sw);
	priv.hshift = lwlib_Log2(sh);

	initDefaultAttr();
}

void Img::expandBits(int bpp)
{
	byte *data;

	assert(priv.data != NULL);
	assert(priv.bpp == 3 || priv.bpp == 4);
	assert(priv.w > 0 && priv.h > 0);
	assert(!priv.mipmap.built);
	assert(!priv.sub.built);

	if (priv.bpp == bpp)
	{
		return;
	}
	assert(priv.bpp == 3);

	const int size = priv.w * priv.h * bpp;
	data = lwlib_CallocMany(byte, size);
	memset(data, 0xff, size);
	reBpp(data, bpp, priv.data, priv.w, priv.h, priv.bpp);

	if (priv.dataOwned)
	{
		free(priv.data);
	}
	priv.data = data;
	priv.bpp = bpp;
	priv.pitch = priv.w * bpp;
	priv.dataOwned = true;

	initDefaultAttr();
}

void Img::swapRGB()
{
	int i;
	byte *data;

	assert(priv.data != NULL);
	assert(priv.bpp == 3 || priv.bpp == 4);
	assert(priv.w > 0 && priv.h > 0);
	assert(!priv.mipmap.built);
	assert(!priv.sub.built);

	data = priv.data;
	for (i = 0; i < priv.w * priv.h; i++)
	{
		lwlib_Swap(data[0], data[2], lwlib_Byte_t);
		data += priv.bpp;
	}
}

void Img::resample(byte *dst, int outwidth, int outheight,
	byte *data, int width, int height, int bpp)
{
	int x, y, n;
	float sx, sy;
	float *tempin;

	// allocate storage for intermediate image
	tempin = (float *)calloc(width * height * bpp, sizeof(float));
	// convert the pixel data to floating point
	for (n = 0; n < width * height * bpp; n++)
		tempin[n] = (float)data[n];

	// Scale the image
	if (outwidth > 1)
		sx = (float)(width - 1) / (float)(outwidth - 1);
	else
		sx = (float)(width - 1);
	if (outheight > 1)
		sy = (float)(height - 1) / (float)(outheight - 1);
	else
		sy = (float)(height - 1);

	if (sx <= 1.0 && sy <= 1.0)
	{
		// magnify both width and height:  use weighted sample of 4 pixels
		int x0, x1, y0, y1;
		float alpha, beta;
		float *src00, *src01, *src10, *src11;
		float s1, s2;

		for (y = 0; y < outheight; y++)
		{
			y0 = (int)(y * sy);
			y1 = y0 + 1;
			if (y1 >= height)
				y1 = height - 1;
			alpha = y * sy - y0;
			y0 *= width;
			y1 *= width;
			for (x = 0; x < outwidth; x++)
			{
				x0 = (int)(x * sx);
				x1 = x0 + 1;
				if (x1 >= width)
					x1 = width - 1;
				beta = x * sx - x0;

				// compute weighted average of pixels in rect (x0, y0) - (x1, y1)
				src00 = tempin + (y0 + x0) * bpp;
				src01 = tempin + (y0 + x1) * bpp;
				src10 = tempin + (y1 + x0) * bpp;
				src11 = tempin + (y1 + x1) * bpp;

				for (n = 0; n < bpp; n++)
				{
					s1 = (*src00++) * (1.0f - beta) + (*src01++) * beta;
					s2 = (*src10++) * (1.0f - beta) + (*src11++) * beta;
					*dst++ = (byte) (s1 * (1.0f - alpha) + s2 * alpha);
				}
			}
		}
	}
	else
	{
		// shrink width and/or height:  use an unweighted box filter
		int x0, x1, y0, y1;
		int xx, yy;
		float sum;

		for (y = 0; y < outheight; y++)
		{
			y0 = (int)(y * sy);
			y1 = y0 + 1;
			if (y1 >= height)
				y1 = height - 1;

			for (x = 0; x < outwidth; x++)
			{
				x0 = (int)(x * sx);
				x1 = x0 + 1;
				if (x1 >= width)
					x1 = width - 1;

				// compute average of pixels in the rectangle (x0, y0) - (x1, y1)
				for (n = 0; n < bpp; n++)
				{
					sum = 0;
					for (yy = y0; yy <= y1; yy++)
						for (xx = x0; xx <= x1; xx++)
							sum += *(tempin + (yy * width + xx) * bpp + n);
					sum /= (x1 - x0 + 1) * (y1 - y0 + 1);
					*dst++ = (byte) sum;
				}
			}
		}
	}

	free(tempin);
}

void Img::resamplePaletted(byte *dst, int outwidth, int outheight,
	byte *data, int width, int height)
{
	int tx, ty;
	lwlib_Frac_t sxfrac, syfrac;

	syfrac = lwlib_FracInit(0, 0, height, outheight);
	for (ty = 0; ty < outheight; ty++)
	{
		sxfrac = lwlib_FracInit(0, 0, width, outwidth);
		for (tx = 0; tx < outwidth; tx++)
		{
			*dst++ = data[syfrac.val * width + sxfrac.val];
			lwlib_FracInc(&sxfrac);
		}
		lwlib_FracInc(&syfrac);
	}
}

void Img::reBpp(uint8_t *dst, int dstbpp, uint8_t *src, int width, 
	int height, int bpp)
{
	int x, y;
	int minBpp;

	minBpp = lwlib_MIN(bpp, dstbpp);

	for (y = 0; y < height; y++)
	{
		for (x = 0; x < width; x++)
		{
			memcpy(dst, src, minBpp);
			src += bpp;
			dst += dstbpp;
		}
	}
}

bool Img::mipMap(byte *src, int *width, int *height, int bpp)
{
	int nw, nh;
	byte *dst;

	if (*width == 1 && *height == 1)
		return false;

	nw = (*width < 2) ? 1 : *width >> 1;
	nh = (*height < 2) ? 1 : *height >> 1;
	dst = (byte *)calloc(nw * nh * bpp, sizeof(byte));
	Img::resample(dst, nw, nh, src, *width, *height, bpp);
	memcpy(src, dst, nw * nh * bpp);
	free(dst);
	*width = nw;
	*height = nh;

	return true;
}

void Img_Modulate(Img_t img, Point4b colorRgba)
{
}
