/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include "lw_keys.h"
#include "lw_timer.h"
#include "lw_strmap.h"
#include "lw_intmap.h"

#define lwlib_Keys_Debug 0
#define lwlib_Keys_DebugChangeState 1
#define lwlib_Keys_DebugSetEvent 1
#define lwlib_Keys_DebugGotKey 1
#define lwlib_Keys_DebugBtnStr "a"

#define lwlib_Keys_MaxSdlKeysPerButton 2
#define lwlib_Keys_RepeatFirstTimeout 500
#define lwlib_Keys_RepeatSecondTimeout 50
#define lwlib_Keys_FastRepeatTimeout 50

#define lwlib_Keys_ButtonType(name) KEYS_BUTTONTYPE_ ## name
#define lwlib_Keys_ButtonState(name) KEYS_BUTTONSTATE_ ## name

typedef enum lwlib_Keys_ButtonType_e
{
	lwlib_Keys_ButtonType(Raw),
	lwlib_Keys_ButtonType(NoMod),
	lwlib_Keys_ButtonType(Mod),
	lwlib_Keys_ButtonType(Last) = lwlib_Keys_ButtonType(Mod),
} lwlib_Keys_ButtonType_t;

typedef enum lwlib_Keys_ButtonState_e
{
	lwlib_Keys_ButtonState(Idle),
	lwlib_Keys_ButtonState(Pressed),
	lwlib_Keys_ButtonState(WaitModsRelease),
	lwlib_Keys_ButtonState(KeyDown_WaitModPress),
	lwlib_Keys_ButtonState(KeyDown_WaitModsRelease),
	lwlib_Keys_ButtonState(ModDown_WaitKeyPress),
	lwlib_Keys_ButtonState(Last) = lwlib_Keys_ButtonState(ModDown_WaitKeyPress),
} lwlib_Keys_ButtonState_t;

typedef struct lwlib_Keys_Button_s
{
	lwlib_StrMapKey_t btnStr;
	lwlib_Keys_ButtonType_t btnType;
	SDLKey sdlKeySym;
	SDLKey sdlKeyMod;
	int eventMask;
	lwlib::Timer repeatTimer;
	int repeatCount;
	lwlib::Timer fastRepeatTimer;
	int btnState;
	int modsDown;
} lwlib_Keys_Button_t;

typedef struct lwlib_Keys_FilterDualKey_s
{
	bool ignoreEvent;
	SDLKey sdlKeySym;
	bool leftDualKeyDown;
	bool rightDualKeyDown;
} lwlib_Keys_FilterDualKey_t;

typedef struct lwlib_Keys_EventQueue_s
{
	lwlib_StrSeq_t btnStrs;
	lwlib_Keys_Event_t event;
	int btnStrNum;
} lwlib_Keys_EventQueue_t;

typedef struct lwlib_Keys_Priv_s
{
	lwlib_StrMap_t buttonMap;
	lwlib_Keys_FilterDualKey_t filterDualKey[3];
	lwlib_IntMap_t sdlKeySymToStringMap;
	bool ignore;
	lwlib_Keys_EventQueue_t eventQueue;
	lwlib_StrSeq_t numericBtnStrs;
	lwlib_StrSeq_t asciiBtnStrs;
	lwlib_StrMap_t btnStrToAsciiMap;
} lwlib_Keys_Priv_t;

lwlib_EnumStr
(
	lwlib_Keys_ButtonType,
	"Raw",
	"NoMod",
	"Mod",
);

lwlib_EnumStr
(
	lwlib_Keys_ButtonState,
	"Idle",
	"Pressed",
	"WaitModsRelease",
	"KeyDown_WaitModPress",
	"KeyDown_WaitModsRelease",
	"ModDown_WaitKeyPress",
);

lwlib_EnumStr
(
	lwlib_Keys_Event,
	"Down",
	"Pressed",
	"Released",
	"Repeat",
	"FastRepeat"
);

static bool s_eventTransient[lwlib_Keys_Event(Last) + 1] = 
{
	/* Down */       false,
	/* Pressed */    true,
	/* Released */   true,
	/* Repeat */     true,
	/* FastRepeat */ true,
};

static void lwlib_Keys_InitSdlKeySymToStringMap(lwlib_Keys_t keys)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	priv->sdlKeySymToStringMap = lwlib_IntMapZero();

	#define RegMap(sym) \
	lwlib_IntMapInsert(&priv->sdlKeySymToStringMap, (int)sym, (lwlib_Word_t)#sym)
	RegMap(SDLK_UNKNOWN); RegMap(SDLK_BACKSPACE); RegMap(SDLK_TAB);
	RegMap(SDLK_CLEAR); RegMap(SDLK_RETURN); RegMap(SDLK_PAUSE); RegMap(SDLK_ESCAPE);
	RegMap(SDLK_SPACE); RegMap(SDLK_EXCLAIM); RegMap(SDLK_QUOTEDBL); RegMap(SDLK_HASH);
	RegMap(SDLK_DOLLAR); RegMap(SDLK_AMPERSAND); RegMap(SDLK_QUOTE); RegMap(SDLK_LEFTPAREN);
	RegMap(SDLK_RIGHTPAREN); RegMap(SDLK_ASTERISK); RegMap(SDLK_PLUS); RegMap(SDLK_COMMA);
	RegMap(SDLK_MINUS); RegMap(SDLK_PERIOD); RegMap(SDLK_SLASH); RegMap(SDLK_0);
	RegMap(SDLK_1); RegMap(SDLK_2); RegMap(SDLK_3); RegMap(SDLK_4);
	RegMap(SDLK_5); RegMap(SDLK_6); RegMap(SDLK_7); RegMap(SDLK_8);
	RegMap(SDLK_9); RegMap(SDLK_COLON); RegMap(SDLK_SEMICOLON); RegMap(SDLK_LESS);
	RegMap(SDLK_EQUALS); RegMap(SDLK_GREATER); RegMap(SDLK_QUESTION); RegMap(SDLK_AT);
	RegMap(SDLK_LEFTBRACKET); RegMap(SDLK_BACKSLASH); RegMap(SDLK_RIGHTBRACKET); RegMap(SDLK_CARET);
	RegMap(SDLK_UNDERSCORE); RegMap(SDLK_BACKQUOTE); RegMap(SDLK_a); RegMap(SDLK_b);
	RegMap(SDLK_c); RegMap(SDLK_d); RegMap(SDLK_e); RegMap(SDLK_f);
	RegMap(SDLK_g); RegMap(SDLK_h); RegMap(SDLK_i); RegMap(SDLK_j);
	RegMap(SDLK_k); RegMap(SDLK_l); RegMap(SDLK_m); RegMap(SDLK_n);
	RegMap(SDLK_o); RegMap(SDLK_p); RegMap(SDLK_q); RegMap(SDLK_r);
	RegMap(SDLK_s); RegMap(SDLK_t); RegMap(SDLK_u); RegMap(SDLK_v);
	RegMap(SDLK_w); RegMap(SDLK_x); RegMap(SDLK_y); RegMap(SDLK_z);
	RegMap(SDLK_DELETE); RegMap(SDLK_KP0); RegMap(SDLK_KP1); RegMap(SDLK_KP2);
	RegMap(SDLK_KP3); RegMap(SDLK_KP4); RegMap(SDLK_KP5); RegMap(SDLK_KP6);
	RegMap(SDLK_KP7); RegMap(SDLK_KP8); RegMap(SDLK_KP9); RegMap(SDLK_KP_PERIOD);
	RegMap(SDLK_KP_DIVIDE); RegMap(SDLK_KP_MULTIPLY); RegMap(SDLK_KP_MINUS); RegMap(SDLK_KP_PLUS);
	RegMap(SDLK_KP_ENTER); RegMap(SDLK_KP_EQUALS); RegMap(SDLK_UP); RegMap(SDLK_DOWN);
	RegMap(SDLK_RIGHT); RegMap(SDLK_LEFT); RegMap(SDLK_INSERT); RegMap(SDLK_HOME);
	RegMap(SDLK_END); RegMap(SDLK_PAGEUP); RegMap(SDLK_PAGEDOWN); RegMap(SDLK_F1);
	RegMap(SDLK_F2); RegMap(SDLK_F3); RegMap(SDLK_F4); RegMap(SDLK_F5);
	RegMap(SDLK_F6); RegMap(SDLK_F7); RegMap(SDLK_F8); RegMap(SDLK_F9);
	RegMap(SDLK_F10); RegMap(SDLK_F11); RegMap(SDLK_F12); RegMap(SDLK_F13);
	RegMap(SDLK_F14); RegMap(SDLK_F15); RegMap(SDLK_NUMLOCK); RegMap(SDLK_CAPSLOCK);
	RegMap(SDLK_SCROLLOCK); RegMap(SDLK_RSHIFT); RegMap(SDLK_LSHIFT); RegMap(SDLK_RCTRL);
	RegMap(SDLK_LCTRL); RegMap(SDLK_RALT); RegMap(SDLK_LALT); RegMap(SDLK_RMETA);
	RegMap(SDLK_LMETA); RegMap(SDLK_LSUPER); RegMap(SDLK_RSUPER); RegMap(SDLK_MODE);
	RegMap(SDLK_COMPOSE); RegMap(SDLK_HELP); RegMap(SDLK_PRINT); RegMap(SDLK_SYSREQ);
	RegMap(SDLK_BREAK); RegMap(SDLK_MENU); RegMap(SDLK_POWER); RegMap(SDLK_EURO);
	RegMap(SDLK_UNDO); RegMap(SDLK_LAST); RegMap(LWLIB_SDLK_MOUSE1); RegMap(LWLIB_SDLK_MOUSE3);
	RegMap(LWLIB_SDLK_MOUSE2); RegMap(LWLIB_SDLK_MWHEELUP); RegMap(LWLIB_SDLK_MWHEELDOWN); RegMap(LWLIB_SDLK_MOUSE4);
	RegMap(LWLIB_SDLK_MOUSE5);
	#undef RegMap
}

static void lwlib_Keys_InitBtnStrs(lwlib_Keys_t keys)
{
	int i, n;
	int keyshift[256];
	char btnChar;
	lwlib_Str_t btnStr, shiftBtnStr;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	lwlib_StrSeqAppend(priv->numericBtnStrs, lwlib_CharLits_0_to_9);

	lwlib_StrSeqAppend(priv->asciiBtnStrs, lwlib_CharLits_0_to_9, lwlib_CharLits_a_to_z, 
		"`", "-", "=", "[", "]", "\\", ";", "'", ",", ".", "/");

	for (n = 0; n < 256; n++)
	{
		keyshift[n] = n;
	}
	for (n = 'a'; n <= 'z'; n++)
	{
		keyshift[n] = n - 'a' + 'A';
	}
	keyshift['1'] = '!';
	keyshift['2'] = '@';
	keyshift['3'] = '#';
	keyshift['4'] = '$';
	keyshift['5'] = '%';
	keyshift['6'] = '^';
	keyshift['7'] = '&';
	keyshift['8'] = '*';
	keyshift['9'] = '(';
	keyshift['0'] = ')';
	keyshift['-'] = '_';
	keyshift['='] = '+';
	keyshift[','] = '<';
	keyshift['.'] = '>';
	keyshift['/'] = '?';
	keyshift[';'] = ':';
	keyshift['\''] = '"';
	keyshift['['] = '{';
	keyshift[']'] = '}';
	keyshift['`'] = '~';
	keyshift['\\'] = '|';
	
	n = vec_size(priv->asciiBtnStrs);
	for (i = 0; i < n; i++)
	{
		btnStr = vec_at(priv->asciiBtnStrs, i);
		btnChar = btnStr.v[0];

		shiftBtnStr = lwlib_Str_Cat(lwlib_Str("shift-"), btnStr);
		vec_push(priv->asciiBtnStrs, lwlib_Str_t, shiftBtnStr);

		lwlib_StrMapInsert(&priv->btnStrToAsciiMap, btnStr.v, (lwlib_Word_t)btnChar);
		lwlib_StrMapInsert(&priv->btnStrToAsciiMap, shiftBtnStr.v, (lwlib_Word_t)keyshift[(int)btnChar]);
	}

	vec_push(priv->asciiBtnStrs, lwlib_Str_t, lwlib_Str("space"));
	lwlib_StrMapInsert(&priv->btnStrToAsciiMap, "space", (lwlib_Word_t)' ');
}

#if lwlib_Keys_Debug
static const char *lwlib_Keys_GetSdlKeySymString(lwlib_Keys_t keys, SDLKey sdlKeySym)
{
	const char *str;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	if (lwlib_IntMapContains(&priv->sdlKeySymToStringMap, (int)sdlKeySym))
	{
		str = (const char *)lwlib_IntMapElem(&priv->sdlKeySymToStringMap, (int)sdlKeySym);
		return lwlib_Va("%s(%d)", str, (int)sdlKeySym);
	}

	return "?";
}
#endif

static void lwlib_Keys_RegisterButton(lwlib_Keys_t keys, lwlib_Keys_Button_t button)
{
	lwlib_Keys_Button_t *newButton;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	assert(!lwlib_StrMapContains(&priv->buttonMap, button.btnStr));

	newButton = lwlib_CallocSingle(lwlib_Keys_Button_t);
	memcpy(newButton, &button, sizeof(lwlib_Keys_Button_t));

	lwlib_StrMapInsert(&priv->buttonMap, button.btnStr, (lwlib_Word_t)newButton);

	assert(lwlib_StrMapContains(&priv->buttonMap, button.btnStr));
}

static void lwlib_Keys_RegisterOneKeyButton(lwlib_Keys_t keys, const char *btnStr, 
	lwlib_Keys_ButtonType_t btnType, SDLKey sdlKeySym, SDLKey sdlKeyMod)
{
	lwlib_Keys_Button_t button;
	lwlib_Zero(button);
	lwlib_StrCopy(button.btnStr, btnStr);
	button.btnType = btnType;
	button.sdlKeySym = sdlKeySym;
	button.sdlKeyMod = sdlKeyMod;
	lwlib_Keys_RegisterButton(keys, button);
}

static void lwlib_Keys_RegisterOneKeyButtonAscii(lwlib_Keys_t keys, 
	lwlib_Keys_ButtonType_t btnType, SDLKey sdlKeySym, SDLKey sdlKeyMod)
{
	const char *fmt;
	lwlib_StrMapKey_t btnStr;
	char btnChar;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	btnChar = (char)sdlKeySym;
	switch (btnType)
	{
	case lwlib_Keys_ButtonType(Raw):
		fmt = "raw-%c";
		break;
	case lwlib_Keys_ButtonType(NoMod):
		fmt = "%c";
		break;
	case lwlib_Keys_ButtonType(Mod):
		switch (sdlKeyMod)
		{
		case SDLK_LSHIFT:
			fmt = "shift-%c";
			break;
		case SDLK_LCTRL:
			fmt = "ctrl-%c";
			break;
		case SDLK_LALT:
			fmt = "alt-%c";
			break;
		default:
			assert(0);
			break;
		}
		break;
	}

	snprintf(btnStr, lwlib_StrMap_MaxString - 1, fmt, btnChar);
	btnStr[lwlib_StrMap_MaxString - 1] = '\0';

	lwlib_Keys_RegisterOneKeyButton(keys, btnStr, btnType, sdlKeySym, sdlKeyMod);
}

static void lwlib_Keys_RegisterOneKeyButtonRangeAscii(lwlib_Keys_t keys,
	lwlib_Keys_ButtonType_t btnType, SDLKey sdlKeySymFirst, SDLKey sdlKeySymLast,
	SDLKey sdlKeyMod)
{
	int sdlKeySym;
	for (sdlKeySym = (int)sdlKeySymFirst; sdlKeySym <= (int)sdlKeySymLast; sdlKeySym++)
	{
		lwlib_Keys_RegisterOneKeyButtonAscii(keys, btnType, (SDLKey)sdlKeySym, sdlKeyMod);
	}
}

static void lwlib_Keys_RegisterAllButtons(lwlib_Keys_t keys)
{
	#define RegKey(btnStr, sdlKeySym) \
	do { \
		lwlib_Keys_RegisterOneKeyButton(keys, btnStr, \
			lwlib_Keys_ButtonType(Raw), sdlKeySym, SDLK_UNKNOWN); \
	} while (0)
	// 1) ctrl, shift, alt
	RegKey("ctrl", SDLK_LCTRL);
	RegKey("shift", SDLK_LSHIFT);
	RegKey("alt", SDLK_LALT);
	#undef RegKey

	#define RegKeyRange(sdlKeySymStart, sdlKeySymLast) \
	do { \
		lwlib_Keys_RegisterOneKeyButtonRangeAscii(keys, \
			lwlib_Keys_ButtonType(Raw), sdlKeySymStart, sdlKeySymLast, SDLK_UNKNOWN); \
		lwlib_Keys_RegisterOneKeyButtonRangeAscii(keys, \
			lwlib_Keys_ButtonType(NoMod), sdlKeySymStart, sdlKeySymLast, SDLK_UNKNOWN); \
		lwlib_Keys_RegisterOneKeyButtonRangeAscii(keys, \
			lwlib_Keys_ButtonType(Mod), sdlKeySymStart, sdlKeySymLast, SDLK_LSHIFT); \
		lwlib_Keys_RegisterOneKeyButtonRangeAscii(keys, \
			lwlib_Keys_ButtonType(Mod), sdlKeySymStart, sdlKeySymLast, SDLK_LCTRL); \
		lwlib_Keys_RegisterOneKeyButtonRangeAscii(keys, \
			lwlib_Keys_ButtonType(Mod), sdlKeySymStart, sdlKeySymLast, SDLK_LALT); \
	} while (0)
	#define RegKey(sdlKeySym) RegKeyRange(sdlKeySym, sdlKeySym)
	// 2) [a-z`0-9-=[]\;',./]
	// 3) shift-[a-z`0-9-=[]\;',./]
	// 4) ctrl-[a-z`0-9-=[]\;',./]
	// 5) alt-[a-z`0-9-=[]\;',./]
	// 6) raw-[a-z`0-9-=[]\;',./]
	RegKeyRange(SDLK_a, SDLK_z);
	RegKey(SDLK_BACKQUOTE);
	RegKeyRange(SDLK_0, SDLK_9);
	RegKey(SDLK_MINUS);
	RegKey(SDLK_EQUALS);
	RegKey(SDLK_LEFTBRACKET);
	RegKey(SDLK_RIGHTBRACKET);
	RegKey(SDLK_BACKSLASH);
	RegKey(SDLK_SEMICOLON);
	RegKey(SDLK_QUOTE);
	RegKey(SDLK_COMMA);
	RegKey(SDLK_PERIOD);
	RegKey(SDLK_SLASH);
	#undef RegKey
	#undef RegKeyRange

	#define RegKey(btnStr, sdlKeySym) \
	do { \
		lwlib_Keys_RegisterOneKeyButton(keys, btnStr, \
			lwlib_Keys_ButtonType(Raw), sdlKeySym, SDLK_UNKNOWN); \
	} while (0)
	// 7) f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12
	RegKey("f1", SDLK_F1);
	RegKey("f2", SDLK_F2);
	RegKey("f3", SDLK_F3);
	RegKey("f4", SDLK_F4);
	RegKey("f5", SDLK_F5);
	RegKey("f6", SDLK_F6);
	RegKey("f7", SDLK_F7);
	RegKey("f8", SDLK_F8);
	RegKey("f9", SDLK_F9);
	RegKey("f10", SDLK_F10);
	RegKey("f11", SDLK_F11);
	RegKey("f12", SDLK_F12);
	// 8) l-ctrl, r-ctrl, l-shift, r-shift, l-alt, r-alt
	RegKey("l-ctrl", SDLK_LCTRL);
	RegKey("r-ctrl", SDLK_RCTRL);
	RegKey("l-shift", SDLK_LSHIFT);
	RegKey("r-shift", SDLK_RSHIFT);
	RegKey("l-alt", SDLK_LALT);
	RegKey("r-alt", SDLK_RALT);
	// 9) enter, space, bksp, del, 
	RegKey("enter", SDLK_RETURN);
	RegKey("space", SDLK_SPACE);
	RegKey("bksp", SDLK_BACKSPACE);
	RegKey("del", SDLK_DELETE);
	// 10) ins, pgup, pgdn, home, end, tab, esc
	RegKey("ins", SDLK_INSERT);
	RegKey("pgup", SDLK_PAGEUP);
	RegKey("pgdn", SDLK_PAGEDOWN);
	RegKey("home", SDLK_HOME);
	RegKey("end", SDLK_END);
	RegKey("tab", SDLK_TAB);
	RegKey("esc", SDLK_ESCAPE);
	// 11) mouse1, mouse2, mouse3, mouse4, mouse5, mwheelup, mwheeldown
	RegKey("mouse1", LWLIB_SDLK_MOUSE1);
	RegKey("mouse3", LWLIB_SDLK_MOUSE3);
	RegKey("mouse2", LWLIB_SDLK_MOUSE2);
	RegKey("mwheelup", LWLIB_SDLK_MWHEELUP);
	RegKey("mwheeldown", LWLIB_SDLK_MWHEELDOWN);
	RegKey("mouse4", LWLIB_SDLK_MOUSE4);
	RegKey("mouse5", LWLIB_SDLK_MOUSE5);
	#undef RegKey

	// 12) left-arrow, right-arrow, up-arrow, down-arrow
	lwlib_Keys_RegisterOneKeyButton(keys, "left-arrow", 
		lwlib_Keys_ButtonType(NoMod), SDLK_LEFT, SDLK_UNKNOWN);
	lwlib_Keys_RegisterOneKeyButton(keys, "right-arrow", 
		lwlib_Keys_ButtonType(NoMod), SDLK_RIGHT, SDLK_UNKNOWN);
	lwlib_Keys_RegisterOneKeyButton(keys, "up-arrow", 
		lwlib_Keys_ButtonType(NoMod), SDLK_UP, SDLK_UNKNOWN);
	lwlib_Keys_RegisterOneKeyButton(keys, "down-arrow", 
		lwlib_Keys_ButtonType(NoMod), SDLK_DOWN, SDLK_UNKNOWN);
	lwlib_Keys_RegisterOneKeyButton(keys, "shift-left-arrow", 
		lwlib_Keys_ButtonType(Mod), SDLK_LEFT, SDLK_LSHIFT);
	lwlib_Keys_RegisterOneKeyButton(keys, "shift-right-arrow", 
		lwlib_Keys_ButtonType(Mod), SDLK_RIGHT, SDLK_LSHIFT);
	lwlib_Keys_RegisterOneKeyButton(keys, "shift-up-arrow", 
		lwlib_Keys_ButtonType(Mod), SDLK_UP, SDLK_LSHIFT);
	lwlib_Keys_RegisterOneKeyButton(keys, "shift-down-arrow", 
		lwlib_Keys_ButtonType(Mod), SDLK_DOWN, SDLK_LSHIFT);
}

static void lwlib_Keys_ClearAllButtons(lwlib_Keys_t keys)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	lwlib_StrMapIterDecl(s);

	lwlib_StrMapIterLoop(priv->buttonMap, s)
	{
		free((lwlib_Keys_Button_t *)s->val);
	}
	lwlib_StrMapClear(&priv->buttonMap);
}

static void lwlib_Keys_FilterDualKeyEvents(lwlib_Keys_t keys, SDLKey sdlKeySym, bool isDown,
	SDLKey leftDualKey, SDLKey rightDualKey, lwlib_Keys_FilterDualKey_t *fdk)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	fdk->sdlKeySym = sdlKeySym;
	if (sdlKeySym == leftDualKey || sdlKeySym == rightDualKey)
	{
		fdk->ignoreEvent = true;
		if (sdlKeySym == leftDualKey && isDown)
		{
			if (fdk->leftDualKeyDown == false && fdk->rightDualKeyDown == false)
			{
				fdk->ignoreEvent = false;
			}
			fdk->leftDualKeyDown = true;
		}
		else if (sdlKeySym == leftDualKey && !isDown)
		{
			if (fdk->leftDualKeyDown == true && fdk->rightDualKeyDown == false)
			{
				fdk->ignoreEvent = false;
			}
			fdk->leftDualKeyDown = false;
		}
		else if (sdlKeySym == rightDualKey && isDown)
		{
			if (fdk->leftDualKeyDown == false && fdk->rightDualKeyDown == false)
			{
				fdk->sdlKeySym = leftDualKey;
				fdk->ignoreEvent = false;
			}
			fdk->rightDualKeyDown = true;
		}
		else if (sdlKeySym == rightDualKey && !isDown)
		{
			if (fdk->leftDualKeyDown == false && fdk->rightDualKeyDown == true)
			{
				fdk->sdlKeySym = leftDualKey;
				fdk->ignoreEvent = false;
			}
			fdk->rightDualKeyDown = false;
		}
	}
}

#if lwlib_Keys_Debug
static void lwlib_Keys_Button_DebugPrint(lwlib_Keys_Button_t *button, const char *format, ...)
{
	va_list argptr;

	if (strcmp(button->btnStr, lwlib_Keys_DebugBtnStr) == 0)
	{
		va_start(argptr, format);
		vprintf(format, argptr);
		va_end(argptr);
	}
}
#endif

static void lwlib_Keys_Button_ChangeState(lwlib_Keys_Button_t *button, int btnState)
{
	#if lwlib_Keys_Debug && lwlib_Keys_DebugChangeState
	lwlib_Keys_Button_DebugPrint(button, "(%s) -> %s\n", button->btnStr, 
		lwlib_Keys_ButtonState_Str[btnState]);
	#endif
	button->btnState = btnState;
}

static void lwlib_Keys_Button_SetEvent(lwlib_Keys_Button_t *button, lwlib_Keys_Event_t event)
{
	#if lwlib_Keys_Debug && lwlib_Keys_DebugSetEvent
	lwlib_Keys_Button_DebugPrint(button, "(%s) %s\n", button->btnStr, 
		lwlib_Keys_Event_Str[event]);
	#endif
	lwlib_SetBitLValue(button->eventMask, event, true);
}

static void lwlib_Keys_Button_ClearEvent(lwlib_Keys_Button_t *button, lwlib_Keys_Event_t event)
{
	lwlib_SetBitLValue(button->eventMask, event, false);
}

static void lwlib_Keys_Button_EnterPressedState(lwlib_Keys_Button_t *button, int btnState)
{
	lwlib_Keys_Button_SetEvent(button, lwlib_Keys_Event(Down));
	lwlib_Keys_Button_SetEvent(button, lwlib_Keys_Event(Pressed));
	lwlib_Keys_Button_SetEvent(button, lwlib_Keys_Event(Repeat));
	lwlib_Keys_Button_SetEvent(button, lwlib_Keys_Event(FastRepeat));
	button->repeatTimer = lwlib::Timer();
	button->repeatCount = 0;
	button->fastRepeatTimer = lwlib::Timer();
	lwlib_Keys_Button_ChangeState(button, btnState);
}

static void lwlib_Keys_Button_ExitPressedState(lwlib_Keys_Button_t *button, int btnState)
{
	lwlib_Keys_Button_ClearEvent(button, lwlib_Keys_Event(Down));
	lwlib_Keys_Button_SetEvent(button, lwlib_Keys_Event(Released));
	lwlib_Keys_Button_ChangeState(button, btnState);
}

static void lwlib_Keys_Button_ProcessRepeat(lwlib_Keys_Button_t *button)
{
	if 
		(
			(
				button->repeatCount == 0 &&
				button->repeatTimer.elapsed(lwlib_Keys_RepeatFirstTimeout)
			) ||
			(
				button->repeatCount > 0 && 
				button->repeatTimer.elapsed(lwlib_Keys_RepeatSecondTimeout)
			)
		)
	{
		lwlib_Keys_Button_SetEvent(button, lwlib_Keys_Event(Repeat));
		button->repeatTimer = lwlib::Timer();
		button->repeatCount++;
	}
	else if (button->fastRepeatTimer.elapsed(lwlib_Keys_FastRepeatTimeout))
	{
		lwlib_Keys_Button_SetEvent(button, lwlib_Keys_Event(FastRepeat));
		button->fastRepeatTimer = lwlib::Timer();
	}
}

static bool lwlib_Keys_KeyIsMod(SDLKey sdlKeySym)
{
	return sdlKeySym == SDLK_LSHIFT || sdlKeySym == SDLK_LCTRL || 
		sdlKeySym == SDLK_LALT;
}

static int lwlib_Keys_ModKeyIndex(SDLKey sdlKeySym)
{
	int modKeyIndex; 
	modKeyIndex =
		(
			sdlKeySym == SDLK_LSHIFT ? 0 : 
			(
				sdlKeySym == SDLK_LCTRL ? 1 : 
				(
					sdlKeySym == SDLK_LALT ? 2 :
					-1
				)
			)
		);
	assert(modKeyIndex != -1);
	return modKeyIndex;
}

static bool lwlib_Keys_ModKeyDown(lwlib_Keys_Button_t *button)
{
	return lwlib_IntegerIsPow2(button->modsDown) && 
		lwlib_BitOn(button->modsDown, lwlib_Keys_ModKeyIndex(button->sdlKeyMod));
}

static void lwlib_Keys_FeedButtonEvent(lwlib_Keys_t keys, lwlib_Keys_Button_t *button,
	SDLKey sdlKeySym, bool isDown)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	if (button->btnType == lwlib_Keys_ButtonType(Raw))
	{
		switch (button->btnState)
		{
		case lwlib_Keys_ButtonState(Idle):
			if (isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_EnterPressedState(button, lwlib_Keys_ButtonState(Pressed));
			}
			break;
		case lwlib_Keys_ButtonState(Pressed):
			if (!isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_ExitPressedState(button, lwlib_Keys_ButtonState(Idle));
			}
			else if (sdlKeySym == SDLK_UNKNOWN)
			{
				lwlib_Keys_Button_ProcessRepeat(button);
			}
			break;
		}
	}
	else if (button->btnType == lwlib_Keys_ButtonType(NoMod))
	{
		switch (button->btnState)
		{
		case lwlib_Keys_ButtonState(Idle):
			if (isDown && lwlib_Keys_KeyIsMod(sdlKeySym))
			{
				lwlib_SetBitLValue(button->modsDown, lwlib_Keys_ModKeyIndex(sdlKeySym), true);
				lwlib_Keys_Button_ChangeState(button, lwlib_Keys_ButtonState(WaitModsRelease));
			}
			else if (isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_EnterPressedState(button, lwlib_Keys_ButtonState(Pressed));
			}
			break;
		case lwlib_Keys_ButtonState(Pressed):
			if (isDown && lwlib_Keys_KeyIsMod(sdlKeySym))
			{
				lwlib_SetBitLValue(button->modsDown, lwlib_Keys_ModKeyIndex(sdlKeySym), true);
				lwlib_Keys_Button_ExitPressedState(button, lwlib_Keys_ButtonState(KeyDown_WaitModsRelease));
			}
			else if (!isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_ExitPressedState(button, lwlib_Keys_ButtonState(Idle));
			}
			else if (sdlKeySym == SDLK_UNKNOWN)
			{
				lwlib_Keys_Button_ProcessRepeat(button);
			}
			break;
		case lwlib_Keys_ButtonState(WaitModsRelease):
			if (lwlib_Keys_KeyIsMod(sdlKeySym))
			{
				lwlib_SetBitLValue(button->modsDown, lwlib_Keys_ModKeyIndex(sdlKeySym), isDown);
				if (button->modsDown == 0)
				{
					lwlib_Keys_Button_ChangeState(button, lwlib_Keys_ButtonState(Idle));
				}
			}
			else if (isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_ChangeState(button, lwlib_Keys_ButtonState(KeyDown_WaitModsRelease));
			}
			break;
		case lwlib_Keys_ButtonState(KeyDown_WaitModsRelease):
			if (lwlib_Keys_KeyIsMod(sdlKeySym))
			{
				lwlib_SetBitLValue(button->modsDown, lwlib_Keys_ModKeyIndex(sdlKeySym), isDown);
				if (button->modsDown == 0)
				{
					lwlib_Keys_Button_EnterPressedState(button, lwlib_Keys_ButtonState(Pressed));
				}
			}
			else if (!isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_ChangeState(button, lwlib_Keys_ButtonState(WaitModsRelease));
			}
			break;
		}
	}
	else if (button->btnType == lwlib_Keys_ButtonType(Mod))
	{
		switch (button->btnState)
		{
		case lwlib_Keys_ButtonState(Idle):
			if (lwlib_Keys_KeyIsMod(sdlKeySym))
			{
				lwlib_SetBitLValue(button->modsDown, lwlib_Keys_ModKeyIndex(sdlKeySym), isDown);
				if (lwlib_Keys_ModKeyDown(button))
				{
					lwlib_Keys_Button_ChangeState(button, lwlib_Keys_ButtonState(ModDown_WaitKeyPress));
				}
			}
			else if (isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_ChangeState(button, lwlib_Keys_ButtonState(KeyDown_WaitModPress));
			}
			break;
		case lwlib_Keys_ButtonState(Pressed):
			if (lwlib_Keys_KeyIsMod(sdlKeySym))
			{
				lwlib_SetBitLValue(button->modsDown, lwlib_Keys_ModKeyIndex(sdlKeySym), isDown);
				if (lwlib_Keys_ModKeyDown(button))
				{
					// no action
				}
				else
				{
					lwlib_Keys_Button_ExitPressedState(button, lwlib_Keys_ButtonState(KeyDown_WaitModPress));
				}
			}
			else if (!isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_ExitPressedState(button, lwlib_Keys_ButtonState(ModDown_WaitKeyPress));
			}
			else if (sdlKeySym == SDLK_UNKNOWN)
			{
				lwlib_Keys_Button_ProcessRepeat(button);
			}
			break;
		case lwlib_Keys_ButtonState(ModDown_WaitKeyPress):
			if (lwlib_Keys_KeyIsMod(sdlKeySym))
			{
				lwlib_SetBitLValue(button->modsDown, lwlib_Keys_ModKeyIndex(sdlKeySym), isDown);
				if (lwlib_Keys_ModKeyDown(button))
				{
					// no action
				}
				else
				{
					lwlib_Keys_Button_ChangeState(button, lwlib_Keys_ButtonState(Idle));
				}
			}
			else if (isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_EnterPressedState(button, lwlib_Keys_ButtonState(Pressed));
			}
			break;
		case lwlib_Keys_ButtonState(KeyDown_WaitModPress):
			if (lwlib_Keys_KeyIsMod(sdlKeySym))
			{
				lwlib_SetBitLValue(button->modsDown, lwlib_Keys_ModKeyIndex(sdlKeySym), isDown);
				if (lwlib_Keys_ModKeyDown(button))
				{
					lwlib_Keys_Button_EnterPressedState(button, lwlib_Keys_ButtonState(Pressed));
				}
			}
			else if (!isDown && sdlKeySym == button->sdlKeySym)
			{
				lwlib_Keys_Button_ChangeState(button, lwlib_Keys_ButtonState(Idle));
			}
			break;
		}
	}
}

static SDLKey lwlib_Keys_FilterDualKey(lwlib_Keys_t keys, SDLKey sdlKeySym, 
	bool isDown, SDLKey modKeyLeft, SDLKey modKeyRight)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	lwlib_Keys_FilterDualKey_t *fdk;

	fdk = &priv->filterDualKey[lwlib_Keys_ModKeyIndex(modKeyLeft)];
	lwlib_Keys_FilterDualKeyEvents(keys, sdlKeySym, isDown,
		modKeyLeft, modKeyRight, fdk);
	if (fdk->ignoreEvent)
	{
		return SDLK_LAST;
	}

	return fdk->sdlKeySym;
}

lwlib_Keys_t lwlib_Keys_New(void)
{
	lwlib_Keys_t keys;
	lwlib_Zero(keys);
	keys.priv = lwlib_CallocSingle(lwlib_Keys_Priv_t);

	lwlib_GetPriv(lwlib_Keys_Priv_t, keys);
	lwlib_Keys_RegisterAllButtons(keys);
	lwlib_Keys_InitSdlKeySymToStringMap(keys);
	lwlib_Keys_InitBtnStrs(keys);

	return keys;
}

lwlib_Keys_t lwlib_Keys_Free(lwlib_Keys_t keys)
{
	lwlib_GetPriv(lwlib_Keys_Priv_t, keys);
	if (priv)
	{
		lwlib_StrMapClear(&priv->btnStrToAsciiMap);
		vec_free(priv->numericBtnStrs, lwlib_Str_t);
		vec_free(priv->asciiBtnStrs, lwlib_Str_t);
		lwlib_IntMapClear(&priv->sdlKeySymToStringMap);
		lwlib_Keys_ClearAllButtons(keys);
		free(priv);
	}
	lwlib_Zero(keys);
	return keys;
}

void lwlib_Keys_ResetTransientEvents(lwlib_Keys_t keys)
{
	int i;
	lwlib_Keys_Button_t *button;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	lwlib_StrMapIterDecl(s);

	lwlib_StrMapIterLoop(priv->buttonMap, s)
	{
		button = (lwlib_Keys_Button_t *)s->val;

		for (i = 0; (size_t)i < lwlib_CountArray(s_eventTransient); i++)
		{
			if (lwlib_BitOn(button->eventMask, i) && s_eventTransient[i])
			{
				lwlib_Keys_Button_ClearEvent(button, (lwlib_Keys_Event_t)i);
			}
		}
	}
}

void lwlib_Keys_ProcessEvent(lwlib_Keys_t keys, SDLKey sdlKeySym, bool isDown)
{
	lwlib_Keys_Button_t *button;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	lwlib_StrMapIterDecl(s);

	sdlKeySym = lwlib_Keys_FilterDualKey(keys, sdlKeySym, isDown,
		SDLK_LSHIFT, SDLK_RSHIFT);
	if (sdlKeySym == SDLK_LAST)
	{
		return;
	}

	sdlKeySym = lwlib_Keys_FilterDualKey(keys, sdlKeySym, isDown,
		SDLK_LCTRL, SDLK_RCTRL);
	if (sdlKeySym == SDLK_LAST)
	{
		return;
	}

	sdlKeySym = lwlib_Keys_FilterDualKey(keys, sdlKeySym, isDown,
		SDLK_LALT, SDLK_RALT);
	if (sdlKeySym == SDLK_LAST)
	{
		return;
	}

	#if lwlib_Keys_Debug && lwlib_Keys_DebugGotKey
	if (sdlKeySym != SDLK_UNKNOWN)
	{
		printf("got %s %s\n", lwlib_Keys_GetSdlKeySymString(keys, sdlKeySym), 
			isDown ? "Down" : "Up");
	}
	#endif

	lwlib_StrMapIterLoop(priv->buttonMap, s)
	{
		button = (lwlib_Keys_Button_t *)s->val;
		lwlib_Keys_FeedButtonEvent(keys, button, sdlKeySym, isDown);
	}
}

bool lwlib_Keys_CheckEvent(lwlib_Keys_t keys, const char *btnStr, lwlib_Keys_Event_t event)
{
	lwlib_Keys_Button_t *button;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	button = (lwlib_Keys_Button_t *)lwlib_StrMapElem(&priv->buttonMap, btnStr);
	if (button == NULL)
	{
		return false;
	}

	return priv->ignore ? false : lwlib_BitOn(button->eventMask, event);
}

bool lwlib_Keys_CheckDown(lwlib_Keys_t keys, const char *btnStr)
{
	return lwlib_Keys_CheckEvent(keys, btnStr, lwlib_Keys_Event(Down));
}

void lwlib_Keys_SetIgnore(lwlib_Keys_t keys, bool ignore)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	priv->ignore = ignore;
}

void lwlib_Keys_EventQueueReset(lwlib_Keys_t keys, lwlib_StrSeq_t btnStrs, lwlib_Keys_Event_t event)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	lwlib_Keys_EventQueue_t *eq = &priv->eventQueue;

	eq->btnStrs = btnStrs;
	eq->event = event;
	eq->btnStrNum = 0;
}

int lwlib_Keys_EventQueuePeek(lwlib_Keys_t keys)
{
	int i;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	lwlib_Keys_EventQueue_t *eq = &priv->eventQueue;

	for (i = eq->btnStrNum; (size_t)i < vec_size(eq->btnStrs); i++)
	{
		if (lwlib_Keys_CheckEvent(keys, vec_at(eq->btnStrs, i).v, eq->event))
		{
			return i;
		}
	}

	return -1;
}

bool lwlib_Keys_EventQueuePending(lwlib_Keys_t keys)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	return lwlib_Keys_EventQueuePeek(keys) != -1;
}

char *lwlib_Keys_EventQueueNext(lwlib_Keys_t keys)
{
	int eqIndex;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	lwlib_Keys_EventQueue_t *eq = &priv->eventQueue;

	eqIndex = lwlib_Keys_EventQueuePeek(keys);
	assert(eqIndex != -1);

	eq->btnStrNum = eqIndex + 1;
	return vec_at(eq->btnStrs, eqIndex).v;
}

char lwlib_Keys_EventQueueNextAscii(lwlib_Keys_t keys)
{
	char *btnStr;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	btnStr = lwlib_Keys_EventQueueNext(keys);
	assert(btnStr != NULL);

	return (char)lwlib_StrMapElem(&priv->btnStrToAsciiMap, btnStr);
}

lwlib_StrSeq_t lwlib_Keys_NumericBtnStrs(lwlib_Keys_t keys)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	return priv->numericBtnStrs;
}

lwlib_StrSeq_t lwlib_Keys_AsciiBtnStrs(lwlib_Keys_t keys)
{
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);
	return priv->asciiBtnStrs;
}

int lwlib_Keys_FirstNumericEvent(lwlib_Keys_t keys, lwlib_Keys_Event_t event)
{
	char *nextKey;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	lwlib_Keys_EventQueueReset(keys, lwlib_Keys_NumericBtnStrs(keys), event);
	while (lwlib_Keys_EventQueuePending(keys))
	{
		nextKey = lwlib_Keys_EventQueueNext(keys);
		assert(nextKey != NULL && *nextKey >= '0' && *nextKey <= '9');
		return *nextKey - '0';
	}

	return -1;
}

void lwlib_Keys_KeyboardFocusReturned(lwlib_Keys_t keys)
{
	int i;
	lwlib_Keys_Button_t *button;
	lwlib_StrMapIterDecl(s);
	lwlib_Keys_FilterDualKey_t *fdk;
	lwlib_GetPrivAssert(lwlib_Keys_Priv_t, keys);

	lwlib_StrMapIterLoop(priv->buttonMap, s)
	{
		button = (lwlib_Keys_Button_t *)s->val;
		button->btnState = lwlib_Keys_ButtonState(Idle);

		for (i = 0; i < lwlib_Keys_Event(Last) + 1; i++)
		{
			lwlib_Keys_Button_ClearEvent(button, (lwlib_Keys_Event_t)i);
		}
	}

	for (i = 0; (size_t)i < lwlib_CountArray(priv->filterDualKey); i++)
	{
		fdk = &priv->filterDualKey[i];
		fdk->ignoreEvent = false;
		fdk->leftDualKeyDown = false;
		fdk->rightDualKeyDown = false;
	}
}
