/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include "lw_mouse.h"
#include "lw_timer.h"
#include "lw_vec.h"
using namespace lwlib;

lwlib_EnumStr(
	lwlib::Mouse_Button,
	"Left",
	"Middle",
	"Right",
);

lwlib_EnumStr(
	lwlib::Mouse_Event,
	"Down",
	"Pressed",
	"Released",
);

Mouse::Mouse(Point2i vidDims)
{
	lwlib_Zero(priv);

	priv.sensitivity = 15.0;
	priv.filter = true;
	priv.vidDims = vidDims;
}

Mouse::~Mouse()
{
	deactivate();
}

void Mouse::setIgnore(bool ignore)
{
	priv.ignore = ignore;
}

void Mouse::activate()
{
	if (!priv.active)
	{
		SDL_ShowCursor(SDL_DISABLE);
		priv.active = true;
	}
}

void Mouse::deactivate()
{
	if (priv.active)
	{
		SDL_ShowCursor(SDL_ENABLE);
		priv.active = false;
	}
}

void Mouse::pollState()
{
	int buttonState;
	int sdlButtons;
	Point2i pos, mPos, mousePos, centerPos;

	if (!priv.active)
	{
		return;
	}

	centerPos = vec_divide(priv.vidDims, 2);

	if (!priv.stuck)
	{
		SDL_WarpMouse(lwlib_ExpandVector2(centerPos));
		sdlButtons = SDL_GetMouseState(&lwlib_X(pos), &lwlib_Y(pos));
		SDL_GetRelativeMouseState(&lwlib_X(pos), &lwlib_Y(pos));
		mPos = vec2i_zero();
		priv.stuck++;
	}
	else
	{
		SDL_GetRelativeMouseState(&lwlib_X(pos), &lwlib_Y(pos));
		sdlButtons = SDL_GetMouseState(&lwlib_X(pos), &lwlib_Y(pos));
		mPos = vec_sub(pos, centerPos);
		SDL_WarpMouse(lwlib_ExpandVector2(centerPos));
	}

	if (priv.filter)
	{
		mousePos = vec_divide(vec_add(mPos, priv.oldMPos), 2);
	}
	else
	{
		mousePos = mPos;
	}
	priv.oldMPos = mPos;

	if (priv.ignore)
	{
		priv.move = vec2i_zero();
	}
	else
	{
		priv.move = vec2f_int(
			vec_scale(vec2i_float(mousePos), priv.sensitivity)
			);
	}

	if (priv.debugMove)
	{
		if (!vec_equal(mPos, vec2i_zero(), 0))
		{
			lwlib_String_t tmpStr[2];
			printf("%s -> %s\n", vec_to_string(tmpStr[0], "%d", mPos), 
				vec_to_string(tmpStr[1], "%d", priv.move));
		}
	}


	buttonState = 0;
	if (sdlButtons & SDL_BUTTON(SDL_BUTTON_LEFT))
	{
		buttonState |= 0x01;
	}

	if (sdlButtons & SDL_BUTTON(SDL_BUTTON_MIDDLE))
	{
		buttonState |= 0x02;
	}

	if (sdlButtons & SDL_BUTTON(SDL_BUTTON_RIGHT))
	{
		buttonState |= 0x04;
	}

	priv.newButtons = buttonState ^ priv.buttonState;
	priv.buttonState = buttonState;
}

void Mouse::addToEdit(Edit *edit)
{
	edit->addItem_bool("filter", &priv.filter);
	edit->addItem_double("sensitivity", &priv.sensitivity);
	edit->addItem_bool("debugMove", &priv.debugMove);
}

Point2i Mouse::getMove()
{
	return priv.move;
}

void Mouse::updateFrame()
{
	if (priv.edit->isActive())
	{
		setActiveFlagState(Mouse_ActiveFlag::EditMode, false);
	}
	else
	{
		setActiveFlagState(Mouse_ActiveFlag::EditMode, true);
	}
}

void Mouse::setEdit(Edit *edit)
{
	priv.edit = edit;
}

bool Mouse::checkButtonStatus(Mouse_Button::e button, 
	Mouse_Event::e event)
{
	bool status;
	int btnMask;

	btnMask = (1 << (int)button);
	status = false;

	switch (event)
	{
	case Mouse_Event::Down:
		status = (priv.buttonState & btnMask) != 0;
		break;
	case Mouse_Event::Pressed:
		status = (priv.newButtons & btnMask) != 0 && 
			(priv.buttonState & btnMask) != 0;
		break;
	case Mouse_Event::Released:
		status = (priv.newButtons & btnMask) != 0 && 
			(priv.buttonState & btnMask) == 0;
		break;
	default:
		break;
	}

	return status;
}

void Mouse::setActiveFlagState(Mouse_ActiveFlag::e activeFlag, bool state)
{
	int old;

	old = priv.activeStatus;
	lwlib_SetBitLValue(priv.activeStatus, activeFlag, state);

	if (priv.activeStatus == old)
	{
		return;
	}

	if (priv.activeStatus == (1 << (int)Mouse_ActiveFlag::Max) - 1)
	{
		activate();
	}
	else
	{
		deactivate();
	}
}
